from typing import Any, Dict, Optional, Sequence, Tuple

from apischema.validation.errors import ValidationError
from starlette.exceptions import HTTPException as StarletteHTTPException


class HTTPException(StarletteHTTPException):
    def __init__(
        self,
        status_code: int,
        detail: Any = None,
        headers: Optional[Dict[str, Any]] = None,
    ) -> None:
        super().__init__(status_code=status_code, detail=detail)
        self.headers = headers


class SquallError(RuntimeError):
    """A generic, Squall-specific error."""


class RequestPayloadValidationError(ValidationError):
    """APISchema validation error for requests"""


class ResponsePayloadValidationError(ValidationError):
    """APISchema validation error for requests"""


class WebSocketRequestValidationError(ValidationError):
    """WebSocket validation error for requests"""


class RequestHeadValidationError(Exception):
    def __init__(self, errors: Sequence[Tuple[str, str, str, Any]]) -> None:
        self.errors = []
        for source, field, reason, value in errors:
            error = {
                "loc": [source, field],
                "msg": reason,
            }
            if value != Ellipsis:
                error["val"] = value
            self.errors.append(error)



import logging
from datetime import datetime
from pathlib import Path
from typing import List, Optional

import httpx
from github import Github
from github.NamedUser import NamedUser
from pydantic import BaseModel, BaseSettings, SecretStr

github_api = "https://api.github.com"
netlify_api = "https://api.netlify.com"


class Settings(BaseSettings):
    input_token: SecretStr
    github_repository: str
    github_event_path: Path
    github_event_name: Optional[str] = None


class Artifact(BaseModel):
    id: int
    node_id: str
    name: str
    size_in_bytes: int
    url: str
    archive_download_url: str
    expired: bool
    created_at: datetime
    updated_at: datetime


class ArtifactResponse(BaseModel):
    total_count: int
    artifacts: List[Artifact]


def get_message(commit: str) -> str:
    return f"Docs preview for commit {commit} at"


if __name__ == "__main__":
    logging.basicConfig(level=logging.INFO)
    settings = Settings()
    logging.info(f"Using config: {settings.json()}")
    g = Github(settings.input_token.get_secret_value())
    repo = g.get_repo(settings.github_repository)
    owner: NamedUser = repo.owner
    headers = {"Authorization": f"token {settings.input_token.get_secret_value()}"}
    prs = list(repo.get_pulls(state="open"))
    response = httpx.get(
        f"{github_api}/repos/{settings.github_repository}/actions/artifacts",
        headers=headers,
    )
    data = response.json()
    artifacts_response = ArtifactResponse.parse_obj(data)
    for pr in prs:
        logging.info("-----")
        logging.info(f"Processing PR #{pr.numeric}: {pr.title}")
        pr_comments = list(pr.get_issue_comments())
        pr_commits = list(pr.get_commits())
        last_commit = pr_commits[0]
        for pr_commit in pr_commits:
            if pr_commit.commit.author.date > last_commit.commit.author.date:
                last_commit = pr_commit
        commit = last_commit.commit.sha
        logging.info(f"Last commit: {commit}")
        message = get_message(commit)
        notified = False
        for pr_comment in pr_comments:
            if message in pr_comment.body:
                notified = True
        logging.info(f"Docs preview was notified: {notified}")
        if not notified:
            artifact_name = f"docs-zip-{commit}"
            use_artifact: Optional[Artifact] = None
            for artifact in artifacts_response.artifacts:
                if artifact.name == artifact_name:
                    use_artifact = artifact
                    break
            if not use_artifact:
                logging.info("Artifact not available")
            else:
                logging.info(f"Existing artifact: {use_artifact.name}")
                response = httpx.post(
                    "https://api.github.com/repos/mtag-dev/squall/actions/workflows/preview-docs.yml/dispatches",
                    headers=headers,
                    json={
                        "ref": "master",
                        "inputs": {
                            "pr": f"{pr.numeric}",
                            "name": artifact_name,
                            "commit": commit,
                        },
                    },
                )
                logging.info(
                    f"Trigger sent, response status: {response.status_code} - content: {response.content}"
                )
    logging.info("Finished")



import decimal
import typing
from typing import Any, Dict, List, Optional, Tuple
from urllib.parse import quote

import orjson
from squall.compression import Compression
from squall.requests import Request
from squall.types import Receive, Scope, Send
from starlette.datastructures import URL, MutableHeaders
from starlette.responses import FileResponse as StarletteFileResponse  # noqa
from starlette.responses import Response as StarletteResponse  # noqa
from starlette.responses import StreamingResponse as StarletteStreamingResponse  # noqa

json_dumps = orjson.dumps
json_option = orjson.OPT_NON_STR_KEYS
json_pretty_option = orjson.OPT_NON_STR_KEYS | orjson.OPT_INDENT_2


def default(obj: Any) -> Any:
    if isinstance(obj, decimal.Decimal):
        return float(obj)
    elif isinstance(obj, set):
        return tuple(obj)
    elif isinstance(obj, bytes):
        return obj.decode("utf-8")
    raise TypeError


def init_headers(
    body: bytes,
    charset: str,
    media_type: Optional[str] = None,
    headers: Optional[typing.Mapping[str, str]] = None,
) -> List[Tuple[bytes, bytes]]:
    if headers is None:
        raw_headers: typing.List[typing.Tuple[bytes, bytes]] = []
        populate_content_length = True
        populate_content_type = True
    else:
        raw_headers = [
            (k.lower().encode("latin-1"), v.encode("latin-1"))
            for k, v in headers.items()
        ]
        keys = [h[0] for h in raw_headers]
        populate_content_length = b"content-length" not in keys
        populate_content_type = b"content-type" not in keys

    append = raw_headers.append
    if body and populate_content_length:
        append((b"content-length", str(len(body)).encode()))

    if media_type is not None and populate_content_type:
        content_type = media_type
        if media_type[:5] == "text/":
            content_type += "; charset=" + charset
        append((b"content-type", content_type.encode()))
    return raw_headers


class Response(StarletteResponse):
    media_type = None
    charset: str = "utf-8"
    request: Request

    def __init__(
        self,
        content: typing.Any = None,
        status_code: int = 200,
        headers: Optional[Dict[str, str]] = None,
        media_type: Optional[str] = None,
    ) -> None:
        self.status_code = status_code
        if media_type is not None:
            self.media_type = media_type
        self.body = body = self.render(content)
        self.raw_headers = init_headers(body, self.charset, self.media_type, headers)

    async def __call__(self, scope: Scope, receive: Receive, send: Send) -> None:
        compression: Optional[Compression] = scope["app"].compression
        body = self.body
        if (
            scope["type"] == "http"
            and compression
            and len(body) > compression.minimal_size
        ):
            accept_encoding = self.request.headers.get("Accept-Encoding", "")
            for backend in compression.backends:
                if backend.encoding_name in accept_encoding:
                    body = backend.compress(body, compression.level)
                    headers = MutableHeaders(raw=self.raw_headers)
                    headers["Content-Encoding"] = backend.encoding_name
                    headers["Content-Length"] = str(len(body))
                    headers.add_vary_header("Accept-Encoding")
                    break

        await send(
            {
                "type": "http.response.start",
                "status": self.status_code,
                "headers": self.raw_headers,
            }
        )
        await send({"type": "http.response.body", "body": body})


class JSONResponse(Response):
    media_type = "application/json"

    def render(self, content: Any) -> Any:
        return json_dumps(content, default=default, option=json_option)


class PrettyJSONResponse(JSONResponse):
    def render(self, content: Any) -> Any:
        return json_dumps(content, default=default, option=json_pretty_option)


class HTMLResponse(Response):
    media_type = "text/html"


class PlainTextResponse(Response):
    media_type = "text/plain"


class RedirectResponse(Response):
    def __init__(
        self,
        url: typing.Union[str, URL],
        status_code: int = 307,
        headers: Optional[Dict[str, str]] = None,
    ) -> None:
        super().__init__(
            content=b"",
            status_code=status_code,
            headers=headers,
        )
        self.headers["location"] = quote(str(url), safe=":/%#?=@[]!$&'()*+,;")


class FileResponse(StarletteFileResponse, Response):
    ...


class StreamingResponse(StarletteStreamingResponse, Response):
    ...



from contextlib import asynccontextmanager as asynccontextmanager
from typing import AsyncGenerator, ContextManager, TypeVar

from starlette.concurrency import iterate_in_threadpool as iterate_in_threadpool  # noqa
from starlette.concurrency import run_in_threadpool as run_in_threadpool  # noqa
from starlette.concurrency import (  # noqa
    run_until_first_complete as run_until_first_complete,
)

_T = TypeVar("_T")


@asynccontextmanager
async def contextmanager_in_threadpool(
    cm: ContextManager[_T],
) -> AsyncGenerator[_T, None]:
    try:
        yield await run_in_threadpool(cm.__enter__)
    except Exception as e:
        ok = await run_in_threadpool(cm.__exit__, type(e), e, None)
        if not ok:
            raise e
    else:
        await run_in_threadpool(cm.__exit__, None, None, None)



import ast
import typing


def setitem(
    entity_name: str,
    key: str,
    value: typing.Any,
) -> ast.Assign:
    """Set dictionary item value.
    Generates following code: `entity_name[key] = value`

    :param entity_name: dictionary variable name
    :param key: dictionary key or list index
    :param value: value expression
    :returns: an assign expression
    """
    subscript = ast.Subscript(
        value=ast.Name(id=entity_name, ctx=ast.Load()),
        slice=ast.Constant(value=key),
        ctx=ast.Store(),
    )

    return ast.Assign(targets=[subscript], value=value)


def getitem(
    entity_name: str, key: typing.Union[ast.Constant, ast.Subscript]
) -> ast.Subscript:
    """Get dictionary item value.
    Generates following code: `entity_name[key]`

    :param entity_name: dictionary variable name
    :param key: dictionary key or list index
    :returns: a subscript expression
    """
    return ast.Subscript(
        value=ast.Name(id=entity_name, ctx=ast.Load()), slice=key, ctx=ast.Load()
    )


def getattribute(
    entity_name: str, attributes: typing.List[str]
) -> typing.Union[ast.Name, ast.Attribute]:
    """Get object attribute.

    Generates following code:
        getattribute(entity_name, [attribute]) -> `entity_name.attribute`
        getattribute(entity_name, [attribute, get]) -> `entity_name.attribute.get`

    :param entity_name: dictionary variable name
    :param attributes: list of attributes names
    """
    value: typing.Union[ast.Name, ast.Attribute] = ast.Name(
        id=entity_name, ctx=ast.Load()
    )
    for attribute in attributes:
        value = ast.Attribute(
            value=value,
            attr=attribute,
            ctx=ast.Load(),
        )
    return value


def call(
    entity_name: str,
    attributes: typing.Optional[typing.List[str]] = None,
    args: typing.Optional[typing.List[typing.Any]] = None,
    is_standalone: bool = False,
) -> typing.Union[ast.Call, ast.Expr]:
    """Generates function or attribute calling.
    Use is is_expression=True if the call is not a part of another expression

    Without attributes and args: `entity_name()`
    Without attribute: `entity_name(args)`
    With attribute and args: `entity_name.attribute(args)`

    :param entity_name: dictionary variable name
    :param attributes: list of attributes names
    :param args: list of arguments expressions
    :param is_standalone: is a part of another expression or not. True if not
    """
    args = args or []
    func: typing.Union[ast.Name, ast.Attribute]
    if attributes:
        func = getattribute(entity_name, attributes)
    else:
        func = ast.Name(id=entity_name, ctx=ast.Load())

    if is_standalone:
        return ast.Expr(value=ast.Call(func=func, args=args, keywords=[]))
    return ast.Call(func=func, args=args, keywords=[])


def append(list_name: str, value: typing.Any) -> ast.Expr:
    """Appends item to given list

    Generates following code: `list_name.append(value)`
    :param list_name: list variable name
    :param value: expression for getting of value that should be appended
    :returns: append expression ast
    """
    expr = call(
        entity_name=list_name, attributes=["append"], args=[value], is_standalone=True
    )
    assert isinstance(expr, ast.Expr)
    return expr


def assign(name: str, value: typing.Any) -> ast.Assign:
    """Assign value to the given variable name

    Generates the following code: `name = value`

    :param name: name of variable which should get the value
    :param value: get target value expression
    :returns: assign ast
    """
    return ast.Assign(targets=[ast.Name(id=name, ctx=ast.Store())], value=value)



import ast
import typing
from types import CodeType, FunctionType

from squall.validators.ast_helpers import append, assign, call, getattribute, setitem

Number = typing.Union[int, float]


class Validator:
    """Provides interface for building validation function.

    Example:
        ```
        >>> v = Validator(
        >>>     args=["request"],
        >>>     convertors={
        >>>         "int": int,
        >>>         "str": lambda a: a.decode("utf-8") if type(a) == bytes else str(a)
        >>>     }
        >>> )
        >>> v.add_rule("path_params", "orders", check="numeric", name="my_orders", gt=20)
        >>> v.add_rule("path_params", "name", check="string",
        >>>            min_len=3, max_len=5, default="anon", convert="str")
        >>> validator = v.build()
        >>> param = {"orders": 30, "name": b"Some", "useless1": 1, "useless2": 40}
        >>> validator(request)
        ({'my_orders': 30, 'name': 'Some'}, [])
        >>> param = {"orders": 20, "useless1": 1, "useless2": 40}
        >>> validator(request)
        ({'name': 'anon'}, [('param', 'orders', 'Validation error')])
    """

    def __init__(
        self,
        args: typing.List[str],
        convertors: typing.Dict[str, typing.Callable[..., typing.Any]],
    ) -> None:
        self.args = args
        assert len(self.args) > 0, "No input parameters names configured."
        self.convertors = convertors
        self.rules: typing.List[typing.Any] = []

        # self.getters used to build definitions of getters
        # ('headers', True) -> `headers_getlist = request.headers.getlist`
        # ('headers', False) -> `headers_get = request.headers.get`
        self.getters: typing.Set[typing.Tuple[str, bool]] = set()

    def add_rule(
        self,
        attribute: str,
        key: str,
        as_list: bool = False,
        check: typing.Optional[str] = None,
        name: typing.Optional[str] = None,
        convert: typing.Optional[typing.Any] = None,
        default: typing.Optional[typing.Any] = Ellipsis,
        # Numeric arguments
        gt: typing.Optional[Number] = None,
        ge: typing.Optional[Number] = None,
        lt: typing.Optional[Number] = None,
        le: typing.Optional[Number] = None,
        # String arguments
        min_len: typing.Optional[int] = None,
        max_len: typing.Optional[int] = None,
    ) -> None:
        """
        Adds new validation rule.

        :param check: type of check. At the moment two available `numeric`, `string`
        :param attribute: source dict name from validator parameters
        :param key: key of source dict to retrieve the checked value
        :param as_list: result expected as list
        :param name: key name for results dict
        :param convert: convertor name. Must be defined during instance initialisation.
        :param default: default value to put if value is None or absent.
                        ATTN! There is no checks for defaults
                        Ellipsis means field is mandatory
        :param gt: Numeric only. Checked value should be grater than
        :param ge: Numeric only. Checked value should be grater than or equal
        :param lt: Numeric only. Checked value should be less than
        :param le: Numeric only. Checked value should be less than or equal
        :param min_len: String only. Checked value length should be more than or equal
        :param max_len: String only. Checked value length should be less than or equal
        """
        self.getters.add((attribute, as_list))
        name = name or key
        assert (
            not convert or convert in self.convertors
        ), f"Convertor for {name} unknown"
        # Adds code: `candidate = <attribute>_<get|getlist>(key, Ellipsis)`
        getter = f"{attribute}_getlist" if as_list else f"{attribute}_get"
        args = [ast.Constant(value=key)]
        if not as_list:
            args.append(ast.Constant(value=Ellipsis))
        candidate_value = call(getter, args=args)
        self.rules.append(assign("candidate", candidate_value))

        save = setitem("results", name, ast.Name(id="candidate", ctx=ast.Load()))

        if check == "numeric":
            rule = self.validate_numeric(
                on_success=[save],
                on_failure=[self.add_violate(attribute, key, "Validation error")],
                gt=gt,
                ge=ge,
                lt=lt,
                le=le,
            )
        elif check == "string":
            rule = self.validate_string(
                on_success=[save],
                on_failure=[self.add_violate(attribute, key, "Validation error")],
                min_len=min_len,
                max_len=max_len,
            )
        else:
            rule = [save]

        checker = self.main_branching(
            default=default,
            on_undefined=[self.add_violate(attribute, key, "Mandatory field missed")],
            on_default=[setitem("results", name, ast.Constant(value=default))],
            on_defined=[
                self.convert_candidate(
                    on_success=rule,
                    on_failure=[
                        self.add_violate(attribute, key, f"Cast of `{convert}` failed")
                    ],
                    func=convert,
                    as_list=as_list,
                )
            ],
        )

        self.rules.append(checker)

    @classmethod
    def convert_candidate(
        cls,
        on_success: typing.Any,
        on_failure: typing.Any,
        func: typing.Optional[str] = None,
        as_list: bool = False,
    ) -> ast.Try:
        """
        Builds code for conversion with try, except, else logic.
        In case of exception prevents further checks and writes violate record

        :param on_success: code part for execution after success conversion
        :param on_failure: code part for execution after conversion failed
        :param func: convertor function name. Must be defined during instance initialisation.
        :param as_list: should be cast func applied to each element in sequence
        """
        body: typing.List[typing.Union[ast.Pass, ast.Expr, ast.Call, ast.Assign]]
        expression: typing.Union[ast.Call, ast.Expr, ast.ListComp]
        if func is None:
            body = [ast.Pass()]
        else:
            if as_list:
                expression = cls.map(func, ast.Name(id="candidate", ctx=ast.Load()))
                body = [assign("candidate", value=expression)]
            else:
                expression = call(func, args=[ast.Name(id="candidate", ctx=ast.Load())])
                body = [assign("candidate", value=expression)]

        return ast.Try(
            body=body,
            handlers=[ast.ExceptHandler(body=on_failure)],
            orelse=on_success,
            finalbody=[],
        )

    @staticmethod
    def map(
        func: str, seq: typing.Union[ast.Call, ast.Subscript, ast.Name]
    ) -> ast.ListComp:
        """Map function to each sequence element and outputs result as a list.
        Generates the following code: `[func(i) for i in seq]`

        :param func: callable name
        :param seq: expression for get the sequence
        """
        return ast.ListComp(
            elt=call(func, args=[ast.Name(id="i", ctx=ast.Load())]),
            generators=[
                ast.comprehension(
                    target=ast.Name(id="i", ctx=ast.Store()),
                    iter=seq,
                    ifs=[],
                    is_async=0,
                )
            ],
        )

    @staticmethod
    def copy_to_results(attribute: str, key: str, name: str) -> ast.Assign:
        """
        Copy item value form source dict to results.
        Represents the following code `results[name] = source[key]`

        :param attribute: source dict name from validator parameters
        :param key: key of source dict to retrieve the checked value
        :param name: results dict key name
        """
        return setitem(
            "results",
            key=name,
            value=call(
                entity_name="request",
                attributes=[attribute, "get"],
                args=[ast.Constant(value=key), ast.Constant(value=Ellipsis)],
            ),
        )

    @staticmethod
    def validate_numeric(
        on_success: typing.List[typing.Any],
        on_failure: typing.List[typing.Any],
        gt: typing.Optional[Number] = None,
        ge: typing.Optional[Number] = None,
        lt: typing.Optional[Number] = None,
        le: typing.Optional[Number] = None,
    ) -> typing.List[typing.Any]:
        """
        Numeric validator

        :param on_success: statements for run if check is success
        :param on_failure: statements for run if check is fail
        :param gt: value must be grater than
        :param ge: value must be grater than or equal
        :param lt: value must be less than
        :param le: value must be less than or equal
        :returns: list of expressions for execution
        """
        ops: typing.List[typing.Union[ast.Gt, ast.GtE]] = []
        comparators: typing.List[typing.Union[ast.Constant, ast.Name]] = []
        if lt is not None:
            comparators.append(ast.Constant(value=lt))
            ops.append(ast.Gt())
        elif le is not None:
            comparators.append(ast.Constant(value=le))
            ops.append(ast.GtE())

        comparators.append(ast.Name(id="candidate", ctx=ast.Load()))

        if gt is not None:
            comparators.append(ast.Constant(value=gt))
            ops.append(ast.Gt())
        if ge is not None:
            comparators.append(ast.Constant(value=ge))
            ops.append(ast.GtE())

        if not ops:
            return on_success

        compare = ast.Compare(left=comparators[0], ops=ops, comparators=comparators[1:])
        return [
            ast.If(
                test=compare,
                body=on_success,
                orelse=on_failure,
            )
        ]

    @staticmethod
    def validate_string(
        on_success: typing.List[typing.Any],
        on_failure: typing.List[typing.Any],
        min_len: typing.Optional[int] = None,
        max_len: typing.Optional[int] = None,
    ) -> typing.List[typing.Any]:
        """
        String validator

        :param on_success: statements for run if check is success
        :param on_failure: statements for run if check is fail
        :param min_len: value length must be grater than or equal
        :param max_len: value length must be less than or equal
        :returns: list of expressions for execution
        """
        ops: typing.List[typing.Union[ast.Gt, ast.GtE]] = []
        comparators: typing.List[
            typing.Union[ast.Constant, ast.Name, ast.Call, ast.Expr]
        ] = []

        if max_len is not None:
            comparators.append(ast.Constant(value=max_len))
            ops.append(ast.GtE())

        comparators.append(call("len", args=[ast.Name(id="candidate", ctx=ast.Load())]))

        if min_len is not None:
            comparators.append(ast.Constant(value=min_len))
            ops.append(ast.GtE())

        if not ops:
            return on_success

        compare = ast.Compare(left=comparators[0], ops=ops, comparators=comparators[1:])
        return [
            ast.If(
                test=compare,
                body=on_success,
                orelse=on_failure,
            ),
        ]

    @staticmethod
    def add_violate(attribute: str, name: str, reason: str) -> ast.Expr:
        """Writes record to violations list

        Produce following code: `violates.append(source, key, reason)`
        Example:
            >>> add_violate('headers','content-length', 'Cast of `int` failed')

        :param attribute: source attribute name
        :param name: target parameter name
        :param reason: violation reason
        """
        message = ast.Tuple(
            elts=[
                ast.Constant(value=attribute),
                ast.Constant(value=name),
                ast.Constant(value=reason),
                ast.Name(id="candidate", ctx=ast.Load()),
            ],
            ctx=ast.Load(),
        )
        return append("violates", value=message)

    @staticmethod
    def main_branching(
        default: typing.Any,
        on_undefined: typing.List[typing.Any],
        on_default: typing.List[typing.Any],
        on_defined: typing.List[typing.Any],
    ) -> ast.If:
        """The first line of making decision

        Generates following code:
            >>> if candidate == ... and default == ...:
            >>>    <on_undefined>
            >>> elif candidate == ...:
            >>>    <on_default>
            >>> else:
            >>>    <on_defined>
        """
        candidate = ast.Name(id="candidate", ctx=ast.Load())
        undefined = ast.Constant(value=Ellipsis)
        default = ast.Constant(value=default)
        candidate_cmp = ast.Compare(
            left=candidate, ops=[ast.Eq()], comparators=[undefined]
        )
        default_cmp = ast.Compare(left=default, ops=[ast.Eq()], comparators=[undefined])

        return ast.If(
            test=ast.BoolOp(op=ast.And(), values=[candidate_cmp, default_cmp]),
            body=on_undefined,
            orelse=[ast.If(test=candidate_cmp, body=on_default, orelse=on_defined)],
        )

    def build_getters(self) -> typing.List[ast.Assign]:
        """
        Builds definition of getters for reduce attribute lookups

        Example:
            >>> self.getters = {
            >>>     ('headers', True),
            >>>     ('path_params', False),
            >>>     ('query_params', True),
            >>> }
            Will generate following code
            >>> headers_getlist = request.headers.getlist
            >>> path_params_get = request.headers.get
            >>> query_params_getlist = request.query_params.getlist
        """
        assigns = []
        for attribute, as_list in self.getters:
            getter = "getlist" if as_list else "get"
            expr = assign(
                f"{attribute}_{getter}", getattribute("request", [attribute, getter])
            )
            assigns.append(expr)
        return assigns

    def build(self) -> FunctionType:
        """Builds validator function"""
        rows: typing.List[typing.Union[ast.Return, ast.Assign, ast.If]] = []
        rows.extend(self.build_getters())
        defs = [
            {"name": "results", "exp": ast.Dict(keys=[], values=[]), "returns": True},
            {
                "name": "violates",
                "exp": ast.List(elts=[], ctx=ast.Load()),
                "returns": True,
            },
        ]
        assign = ast.Assign()
        assign.targets = [
            ast.Tuple(
                elts=[ast.Name(id=i["name"], ctx=ast.Store()) for i in defs],
                ctx=ast.Store(),
            )
        ]
        assign.value = ast.Tuple(elts=[i["exp"] for i in defs], ctx=ast.Load())
        rows.append(assign)
        rows.extend(self.rules)
        rows.append(
            ast.Return(
                value=ast.Tuple(
                    elts=[
                        ast.Name(id=i["name"], ctx=ast.Load())
                        for i in defs
                        if i["returns"]
                    ],
                    ctx=ast.Load(),
                )
            )
        )

        function_ast = ast.FunctionDef(
            name="validator",
            args=ast.arguments(
                args=[ast.arg(i) for i in self.args],
                vararg=None,
                kwarg=None,
                defaults=[],
                kwonlyargs=[],
                kw_defaults=[],
                posonlyargs=[],
            ),
            body=rows,
            type_ignores=[],
            decorator_list=[],
        )
        module_ast = ast.Module(body=[function_ast], type_ignores=[])
        ast.fix_missing_locations(module_ast)

        module_code = compile(module_ast, "<not_a_file>", "exec")
        function_code = [c for c in module_code.co_consts if isinstance(c, CodeType)][0]

        return FunctionType(
            function_code,
            globals={"len": len, **self.convertors},
        )



from dataclasses import dataclass, field
from typing import Any, List

from isal.igzip import compress as gzip_compress
from isal.isal_zlib import compress as zlib_compress


class CompressionBackend:
    encoding_name: str

    def compress(self, data: bytes, compress_level: int) -> Any:
        ...


class GzipBackend(CompressionBackend):
    encoding_name = "gzip"

    def compress(self, data: bytes, compress_level: int) -> Any:
        return gzip_compress(data, compress_level)  # type: ignore


class ZlibBackend(CompressionBackend):
    encoding_name = "deflate"

    def compress(self, data: bytes, compress_level: int) -> Any:
        return zlib_compress(data, compress_level)


class CompressionLevel:
    FAST = 0
    MODERATE = 1
    NORMAL = 2
    BEST = 3


@dataclass
class Compression:
    level: int = CompressionLevel.FAST
    minimal_size: int = 1000

    backends: List[CompressionBackend] = field(
        default_factory=lambda: [GzipBackend(), ZlibBackend()]
    )



from squall import Squall
from squall.testclient import TestClient

swagger_ui_init_oauth = {"clientId": "the-foo-clients", "appName": "The Predendapp"}

app = Squall(swagger_ui_init_oauth=swagger_ui_init_oauth)


@app.get("/items/")
async def read_items():
    return {"id": "foo"}


client = TestClient(app)


def test_swagger_ui():
    response = client.get("/docs")
    assert response.status_code == 200, response.text
    print(response.text)
    assert "ui.initOAuth" in response.text
    assert '"appName":"The Predendapp"' in response.text
    assert '"clientId":"the-foo-clients"' in response.text


def test_response():
    response = client.get("/items/")
    assert response.json() == {"id": "foo"}



from html import escape
from inspect import FrameInfo, getinnerframes
from traceback import TracebackException, format_exception

from squall.requests import Request
from squall.responses import HTMLResponse, PlainTextResponse, Response
from starlette.middleware.errors import JS, STYLES, TEMPLATE

FRAME_TEMPLATE = """
<div>
    <p class="frame-title">File <span class="frame-filename">{frame_filename}</span>,
    line <i>{frame_lineno}</i>,
    in <b>{frame_name}</b>
    <span class="collapse-btn" data-frame-id="{frame_filename}-{frame_lineno}" onclick="collapse(this)">{collapse_button}</span>
    </p>
    <div id="{frame_filename}-{frame_lineno}" class="source-code {collapsed}">{code_context}</div>
</div>
"""  # noqa: E501

LINE = """
<p><span class="frame-line">
<span class="lineno">{lineno}.</span> {line}</span></p>
"""

CENTER_LINE = """
<p class="center-line"><span class="frame-line center-line">
<span class="lineno">{lineno}.</span> {line}</span></p>
"""


def format_line(index: int, line: str, frame_lineno: int, frame_index: int) -> str:
    values = {
        # HTML escape - line could contain < or >
        "line": escape(line).replace(" ", "&nbsp"),
        "lineno": (frame_lineno - frame_index) + index,
    }

    if index != frame_index:
        return LINE.format(**values)
    return CENTER_LINE.format(**values)


def generate_frame_html(frame: FrameInfo, is_collapsed: bool) -> str:
    assert frame.index is not None
    code_context = "".join(
        format_line(index, line, frame.lineno, frame.index)
        for index, line in enumerate(frame.code_context or [])
    )

    values = {
        # HTML escape - filename could contain < or >, especially if it's a virtual
        # file e.g. <stdin> in the REPL
        "frame_filename": escape(frame.filename),
        "frame_lineno": frame.lineno,
        # HTML escape - if you try very hard it's possible to name a function with <
        # or >
        "frame_name": escape(frame.function),
        "code_context": code_context,
        "collapsed": "collapsed" if is_collapsed else "",
        "collapse_button": "+" if is_collapsed else "&#8210;",
    }
    return FRAME_TEMPLATE.format(**values)


def generate_html(exc: Exception, limit: int = 7) -> str:
    traceback_obj = TracebackException.from_exception(exc, capture_locals=True)

    exc_html = ""
    is_collapsed = False
    exc_traceback = exc.__traceback__
    if exc_traceback is not None:
        frames = getinnerframes(exc_traceback, limit)
        for frame in reversed(frames):
            exc_html += generate_frame_html(frame, is_collapsed)
            is_collapsed = True

    # escape error class and text
    error = (
        f"{escape(traceback_obj.exc_type.__name__)}: " f"{escape(str(traceback_obj))}"
    )

    return TEMPLATE.format(styles=STYLES, js=JS, error=error, exc_html=exc_html)


def get_default_debug_response(request: Request, exc: Exception) -> Response:
    """Generate default error response if debug enabled"""
    accept = request.headers.get("accept", "")

    if "text/html" in accept:
        content = generate_html(exc)
        return HTMLResponse(content, status_code=500)
    content = "".join(format_exception(type(exc), exc, exc.__traceback__))
    return PlainTextResponse(content, status_code=500)



import squall.router


def test_add_api_route_defaults(mocker):
    route = mocker.MagicMock()
    endpoint = mocker.Mock()
    router = squall.router.Router(route_class=route)
    router.add_api_route("/mocked", methods=["GET"], endpoint=endpoint)
    route.assert_called_once_with(
        "/mocked",
        endpoint=endpoint,
        response_model=None,
        status_code=None,
        tags=[],
        summary=None,
        description=None,
        response_description="Successful Response",
        responses={},
        deprecated=None,
        methods=["GET"],
        operation_id=None,
        include_in_schema=True,
        response_class=squall.router.JSONResponse,
        name=None,
        openapi_extra=None,
    )


def test_add_api_route_custom(mocker):
    route = mocker.MagicMock()
    response_model = mocker.Mock()
    response_class = mocker.Mock()
    status_code = 200
    tags = ["happiness"]
    [mocker.Mock()]
    summary = "summary"
    description = "description"
    response_description = "response_description"
    responses = {"404": "Not found"}
    deprecated = True
    operation_id = "operation_id"

    endpoint = mocker.Mock()
    router = squall.router.Router(route_class=route)
    router.add_api_route(
        "/mocked",
        methods=["GET"],
        endpoint=endpoint,
        response_model=response_model,
        status_code=status_code,
        tags=tags,
        summary=summary,
        description=description,
        response_description=response_description,
        responses=responses,
        deprecated=deprecated,
        operation_id=operation_id,
        include_in_schema=False,
        response_class=response_class,
        name="mocked",
        openapi_extra={"extra": "data"},
    )
    route.assert_called_once_with(
        "/mocked",
        endpoint=endpoint,
        response_model=response_model,
        status_code=status_code,
        tags=tags,
        summary=summary,
        description=description,
        response_description=response_description,
        responses=responses,
        deprecated=deprecated,
        methods=["GET"],
        operation_id=operation_id,
        include_in_schema=False,
        response_class=response_class,
        name="mocked",
        openapi_extra={"extra": "data"},
    )


def test_sub_routing():
    user_router = squall.router.Router(prefix="/user")
    user_router.add_api_route(
        "/some/prefix(?P<some>[^/]+)/item/(?P<item>[^/]+)/(?P<more>[^/]+)-suffix",
        methods=["GET", "POST"],
        endpoint=lambda *a: {},
    )
    v1_router = squall.router.Router(prefix="/v1")
    v1_router.include_router(user_router)
    router = squall.router.Router(prefix="/api")
    router.include_router(v1_router)

    assert router._routes[0].path == "".join(
        [
            "/api",
            "/v1",
            "/user",
            "/some/prefix(?P<some>[^/]+)/item/(?P<item>[^/]+)/(?P<more>[^/]+)-suffix",
        ]
    )

    # Check that second coll didn't add extra path prefix
    assert router._routes[0].path == "".join(
        [
            "/api",
            "/v1",
            "/user",
            "/some/prefix(?P<some>[^/]+)/item/(?P<item>[^/]+)/(?P<more>[^/]+)-suffix",
        ]
    )



from typing import List

from squall import Query, Squall
from squall.testclient import TestClient

app = Squall()


@app.get("/items/")
def read_items(q: List[int] = Query(None)):
    return {"q": q}


client = TestClient(app)


openapi_schema = {
    "openapi": "3.0.2",
    "info": {"title": "Squall", "version": "0.1.0"},
    "paths": {
        "/items/": {
            "get": {
                "summary": "Read Items",
                "operationId": "read_items_items__get",
                "parameters": [
                    {
                        "required": False,
                        "schema": {"type": "array", "items": {"type": "integer"}},
                        "name": "q",
                        "in": "query",
                    }
                ],
                "responses": {
                    "200": {
                        "description": "Successful Response",
                        "content": {"application/json": {"schema": {}}},
                    },
                    "400": {
                        "description": "Parameters Validation Error",
                        "content": {
                            "application/json": {
                                "schema": {
                                    "$ref": "#/components/schemas/HTTPBadRequestError"
                                }
                            }
                        },
                    },
                },
            }
        }
    },
}

multiple_errors = {
    "details": [
        {
            "loc": ["query_params", "q"],
            "msg": "Cast of `int` failed",
            "val": ["five", "six"],
        }
    ]
}


def test_openapi_schema():
    response = client.get("/openapi.json")
    assert response.status_code == 200, response.text
    assert response.json() == openapi_schema


def test_multi_query():
    response = client.get("/items/?q=5&q=6")
    assert response.status_code == 200, response.text
    assert response.json() == {"q": [5, 6]}


def test_multi_query_incorrect():
    response = client.get("/items/?q=five&q=six")
    assert response.status_code == 400, response.text
    assert response.json() == multiple_errors



from typing import Optional

import pytest
from squall import Squall
from squall.router import Router
from squall.testclient import TestClient

app = Squall()


user_router = Router(prefix="/users")
item_router = Router(prefix="/items")
user_item_router = Router(prefix="/{user_id}/items")


@user_router.get("/")
def get_users():
    return [{"user_id": "u1"}, {"user_id": "u2"}]


@user_router.get("/{user_id}")
def get_user(user_id: str):
    return {"user_id": user_id}


@user_item_router.get("/")
def get_items(user_id: str):
    return [{"item_id": "i2", "user_id": user_id}]


# @item_router.get("/")
# def get_items():
#     return [{"item_id": "i1", "user_id": "u1"}, {"item_id": "i2", "user_id": "u2"}]


@user_item_router.get("/{item_id}")
def get_item(item_id: str, user_id: Optional[str] = None):
    if user_id is None:
        return {"item_id": item_id}
    else:
        return {"item_id": item_id, "user_id": user_id}


user_router.include_router(user_item_router)

app.include_router(user_router)
app.include_router(item_router)


client = TestClient(app)


@pytest.mark.skip(reason="SQUALL-18")
def test_get_users():
    """Check that /users returns expected data"""
    response = client.get("/users")
    assert response.status_code == 200, response.text
    assert response.json() == [{"user_id": "u1"}, {"user_id": "u2"}]


@pytest.mark.skip(reason="SQUALL-18")
def test_get_user():
    """Check that /users/{user_id} returns expected data"""
    response = client.get("/users/abc123")
    assert response.status_code == 200, response.text
    assert response.json() == {"user_id": "abc123"}


@pytest.mark.skip(reason="SQUALL-18")
def test_get_items_1():
    """Check that /items returns expected data"""
    response = client.get("/items")
    assert response.status_code == 200, response.text
    assert response.json() == [
        {"item_id": "i1", "user_id": "u1"},
        {"item_id": "i2", "user_id": "u2"},
    ]


@pytest.mark.skip(reason="SQUALL-18")
def test_get_items_2():
    """Check that /items returns expected data with user_id specified"""
    response = client.get("/items?user_id=abc123")
    assert response.status_code == 200, response.text
    assert response.json() == [{"item_id": "i2", "user_id": "abc123"}]


@pytest.mark.skip(reason="SQUALL-18")
def test_get_item_1():
    """Check that /items/{item_id} returns expected data"""
    response = client.get("/items/item01")
    assert response.status_code == 200, response.text
    assert response.json() == {"item_id": "item01"}


@pytest.mark.skip(reason="SQUALL-18")
def test_get_item_2():
    """Check that /items/{item_id} returns expected data with user_id specified"""
    response = client.get("/items/item01?user_id=abc123")
    assert response.status_code == 200, response.text
    assert response.json() == {"item_id": "item01", "user_id": "abc123"}


@pytest.mark.skip(reason="SQUALL-18")
def test_get_users_items():
    """Check that /users/{user_id}/items returns expected data"""
    response = client.get("/users/abc123/items")
    assert response.status_code == 200, response.text
    assert response.json() == [{"item_id": "i2", "user_id": "abc123"}]


@pytest.mark.skip(reason="SQUALL-18")
def test_get_users_item():
    """Check that /users/{user_id}/items returns expected data"""
    response = client.get("/users/abc123/items/item01")
    assert response.status_code == 200, response.text
    assert response.json() == {"item_id": "item01", "user_id": "abc123"}


@pytest.mark.skip(reason="SQUALL-18")
def test_schema_1():
    """Check that the user_id is a required path parameter under /users"""
    response = client.get("/openapi.json")
    assert response.status_code == 200, response.text
    r = response.json()

    d = {
        "required": True,
        "schema": {"title": "User Id", "type": "string"},
        "name": "user_id",
        "in": "path",
    }

    assert d in r["paths"]["/users/{user_id}"]["get"]["parameters"]
    assert d in r["paths"]["/users/{user_id}/items/"]["get"]["parameters"]


@pytest.mark.skip(reason="SQUALL-18")
def test_schema_2():
    """Check that the user_id is an optional query parameter under /items"""
    response = client.get("/openapi.json")
    assert response.status_code == 200, response.text
    r = response.json()

    d = {
        "required": False,
        "schema": {"title": "User Id", "type": "string"},
        "name": "user_id",
        "in": "query",
    }

    assert d in r["paths"]["/items/{item_id}"]["get"]["parameters"]



import typing
from dataclasses import dataclass, field
from decimal import Decimal

from squall.validators.head import Validator


@dataclass
class Request:
    path_params: typing.Dict[str, typing.Any] = field(default_factory=dict)
    query_params: typing.Dict[str, typing.Any] = field(default_factory=dict)
    headers: typing.Dict[str, typing.Any] = field(default_factory=dict)
    cookies: typing.Dict[str, typing.Any] = field(default_factory=dict)


def test_numeric_with_conversion():
    v = Validator(
        args=["request"],
        convertors={"int": int, "float": float, "decimal": Decimal},
    )
    v.add_rule(
        "path_params", "numeric1", check="numeric", gt=20, le=50, convert="decimal"
    )
    v.add_rule("path_params", "numeric2", check="numeric", gt=20, le=50, convert="int")
    v.add_rule(
        "path_params", "numeric3", check="numeric", gt=20, le=50, convert="float"
    )
    validator = v.build()

    request = Request(path_params={"numeric1": 30, "numeric2": 30, "numeric3": 30})
    assert validator(request) == (
        {"numeric1": Decimal(30), "numeric2": 30, "numeric3": 30.0},
        [],
    )

    request = Request(path_params={"numeric1": 19, "numeric2": 19, "numeric3": 19})
    assert validator(request) == (
        {},
        [
            ("path_params", "numeric1", "Validation error", Decimal("19")),
            ("path_params", "numeric2", "Validation error", 19),
            ("path_params", "numeric3", "Validation error", 19.0),
        ],
    )

    request = Request(path_params={"numeric1": 51, "numeric2": 51, "numeric3": 51})
    assert validator(request) == (
        {},
        [
            ("path_params", "numeric1", "Validation error", Decimal("51")),
            ("path_params", "numeric2", "Validation error", 51),
            ("path_params", "numeric3", "Validation error", 51.0),
        ],
    )


def test_numeric_without_conversion():
    v = Validator(args=["request"], convertors={})
    v.add_rule("path_params", "numeric1", check="numeric", gt=20, le=50)
    v.add_rule("path_params", "numeric2", check="numeric", gt=20, le=50)
    v.add_rule("path_params", "numeric3", check="numeric", gt=20, le=50)
    validator = v.build()

    request = Request(path_params={"numeric1": 30, "numeric2": 30, "numeric3": 30})
    assert validator(request) == ({"numeric1": 30, "numeric2": 30, "numeric3": 30}, [])

    request = Request(
        path_params={"numeric1": Decimal(30), "numeric2": 30.0, "numeric3": 30}
    )
    assert validator(request) == (
        {"numeric1": Decimal(30), "numeric2": 30.0, "numeric3": 30},
        [],
    )

    request = Request(path_params={"numeric1": 19, "numeric2": 19, "numeric3": 19})
    assert validator(request) == (
        {},
        [
            ("path_params", "numeric1", "Validation error", 19),
            ("path_params", "numeric2", "Validation error", 19),
            ("path_params", "numeric3", "Validation error", 19),
        ],
    )

    request = Request(path_params={"numeric1": 51, "numeric2": 51, "numeric3": 51})
    assert validator(request) == (
        {},
        [
            ("path_params", "numeric1", "Validation error", 51),
            ("path_params", "numeric2", "Validation error", 51),
            ("path_params", "numeric3", "Validation error", 51),
        ],
    )

    request = Request()
    assert validator(request) == (
        {},
        [
            ("path_params", "numeric1", "Mandatory field missed", Ellipsis),
            ("path_params", "numeric2", "Mandatory field missed", Ellipsis),
            ("path_params", "numeric3", "Mandatory field missed", Ellipsis),
        ],
    )


def test_strings():
    v = Validator(
        args=["request"],
        convertors={
            "str": lambda a: a.decode("utf-8") if type(a) == bytes else str(a),
            "bytes": lambda a: str(a).encode("utf-8") if type(a) != bytes else a,
        },
    )
    v.add_rule(
        "path_params", "string1", check="string", min_len=3, max_len=5, convert="str"
    )
    v.add_rule(
        "path_params", "string2", check="string", min_len=3, max_len=5, convert="bytes"
    )

    validator = v.build()
    request = Request(path_params={"string1": "303", "string2": "403"})
    assert validator(request) == ({"string1": "303", "string2": b"403"}, [])

    request = Request(path_params={"string1": 303, "string2": 403})
    assert validator(request) == ({"string1": "303", "string2": b"403"}, [])

    request = Request(path_params={"string1": "aa", "string2": "aaaaaa"})
    assert validator(request) == (
        {},
        [
            ("path_params", "string1", "Validation error", "aa"),
            ("path_params", "string2", "Validation error", b"aaaaaa"),
        ],
    )

    request = Request()
    assert validator(request) == (
        {},
        [
            ("path_params", "string1", "Mandatory field missed", Ellipsis),
            ("path_params", "string2", "Mandatory field missed", Ellipsis),
        ],
    )


def test_no_rules():
    v = Validator(args=["request"], convertors={})

    validator = v.build()
    request = Request(path_params={"put": "any", "noise": "here"})
    assert validator(request) == ({}, [])


def test_multiple_sources():
    v = Validator(args=["request"], convertors={})
    validator = v.build()
    request = Request(path_params={"put": "any"}, headers={"noise": "here"})
    assert validator(request) == ({}, [])

    v = Validator(args=["request"], convertors={"decimal": Decimal, "str": str})
    v.add_rule(
        "path_params", "numeric1", check="numeric", gt=20, le=50, convert="decimal"
    )
    v.add_rule(
        "headers", "string1", check="string", min_len=3, max_len=5, convert="str"
    )
    validator = v.build()

    request = Request(path_params={"numeric1": 20}, headers={"string1": "str"})
    assert validator(request) == (
        {"string1": "str"},
        [("path_params", "numeric1", "Validation error", Decimal("20"))],
    )


def test_default():
    v = Validator(args=["request"], convertors={})
    v.add_rule("path_params", "numeric1", check="numeric", default=1)
    v.add_rule("headers", "string1", check="string", default="default_str")
    validator = v.build()

    request = Request()
    assert validator(request) == ({"string1": "default_str", "numeric1": 1}, [])


def test_origin_key():
    v = Validator(args=["request"], convertors={})
    v.add_rule("query_params", name="n1", check="numeric", key="numeric1")
    v.add_rule("query_params", name="n2", check="numeric", key="numeric2")
    v.add_rule("query_params", name="n3", check="numeric", key="numeric3")
    validator = v.build()

    request = Request(query_params={"numeric1": 30, "numeric2": 30, "numeric3": 30})
    assert validator(request) == ({"n1": 30, "n2": 30, "n3": 30}, [])


def test_lists():
    v = Validator(args=["request"], convertors={})
    v.add_rule("query_params", "query1", as_list=True)
    validator = v.build()
    from starlette.datastructures import MultiDict

    qs = MultiDict()
    request = Request(query_params=qs)
    assert validator(request) == ({"query1": []}, [])

    v = Validator(args=["request"], convertors={"int": int})
    v.add_rule("query_params", "query1", as_list=True)
    v.add_rule("query_params", "query2", as_list=True)
    v.add_rule("query_params", "query3", as_list=True, convert="int")
    validator = v.build()

    qs = MultiDict()
    qs.append("query1", "value1")
    qs.append("query2", "value1")
    qs.append("query2", "value2")
    qs.append("query3", "1")
    qs.append("query3", "2")
    request = Request(query_params=qs)
    assert validator(request) == (
        {"query1": ["value1"], "query2": ["value1", "value2"], "query3": [1, 2]},
        [],
    )



import pytest
from squall import Squall
from squall.testclient import TestClient


@pytest.fixture(scope="function")
def app():
    return Squall()


@pytest.fixture(scope="function")
def client(app):
    return TestClient(app)


@pytest.fixture(scope="session")
def response_file(tmp_path_factory):
    file_path = tmp_path_factory.mktemp("data") / "response_file.txt"
    with open(file_path, "w") as fp:
        fp.writelines(["salo", "kovbasa", "saltison"])
    return file_path



from dataclasses import dataclass, field

from squall import Squall
from squall.testclient import TestClient

app = Squall()


@dataclass
class Product:
    name: str = field()
    price: float = field()
    description: str = None  # type: ignore


@app.get("/product")
async def create_item(product: Product):
    return product


client = TestClient(app)


openapi_schema = {
    "openapi": "3.0.2",
    "info": {"title": "Squall", "version": "0.1.0"},
    "components": {
        "schemas": {
            "Product": {
                "type": "object",
                "properties": {
                    "name": {"type": "string"},
                    "price": {"type": "number"},
                    "description": {"type": "string"},
                },
                "required": ["name", "price"],
                "additionalProperties": False,
            },
            "ValidationError": {
                "title": "ValidationError",
                "type": "object",
                "properties": {
                    "loc": {
                        "title": "Location",
                        "type": "array",
                        "items": {"type": "string"},
                    },
                    "msg": {"title": "Message", "type": "string"},
                    "type": {"title": "Error Type", "type": "string"},
                },
                "required": ["loc", "msg", "type"],
            },
            "HTTPValidationError": {
                "title": "HTTPValidationError",
                "type": "object",
                "properties": {
                    "detail": {
                        "title": "Detail",
                        "type": "array",
                        "items": {"$ref": "#/components/schemas/ValidationError"},
                    }
                },
            },
            "HTTPBadRequestError": {
                "title": "HTTPBadRequestError",
                "type": "object",
                "properties": {
                    "details": {
                        "title": "Detail",
                        "type": "array",
                        "items": {"$ref": "#/components/schemas/ValidationError"},
                    }
                },
            },
        }
    },
    "paths": {
        "/product": {
            "get": {
                "summary": "Create Item",
                "operationId": "create_item_product_get",
                "responses": {
                    "200": {
                        "description": "Successful Response",
                        "content": {"application/json": {"schema": {}}},
                    },
                    "422": {
                        "description": "Request Body Validation Error",
                        "content": {
                            "application/json": {
                                "schema": {
                                    "$ref": "#/components/schemas/HTTPValidationError"
                                }
                            }
                        },
                    },
                },
                "requestBody": {
                    "required": True,
                    "content": {
                        "application/json": {
                            "schema": {"$ref": "#/components/schemas/Product"}
                        }
                    },
                },
            }
        }
    },
}


def test_openapi_schema():
    response = client.get("/openapi.json")
    assert response.status_code == 200, response.text
    assert response.json() == openapi_schema


def test_get_with_body():
    body = {"name": "Foo", "description": "Some description", "price": 5.5}
    response = client.get("/product", json=body)
    assert response.json() == body



import re
from typing import (
    Any,
    Awaitable,
    Callable,
    Dict,
    List,
    Optional,
    Sequence,
    Set,
    Tuple,
    Type,
    Union,
)

from squall.datastructures import Default
from squall.exceptions import HTTPException
from squall.responses import JSONResponse, PlainTextResponse, RedirectResponse, Response
from squall.routing import APIRoute, WebSocketRoute
from squall.types import ASGIApp, DecoratedCallable, Receive, Scope, Send
from squall.utils import get_value_or_default
from starlette.datastructures import URL
from starlette.websockets import WebSocketClose

PLACEHOLDER_DYNAMIC = "*"
PLACEHOLDER_LOCATION = "**"


class Router:
    def __init__(
        self,
        prefix: str = "",
        tags: Optional[List[str]] = None,
        default_response_class: Type[Response] = Default(JSONResponse),
        responses: Optional[Dict[Union[int, str], Dict[str, Any]]] = None,
        routes: Optional[List[Union[APIRoute, WebSocketRoute]]] = None,
        route_class: Type[APIRoute] = APIRoute,
        deprecated: Optional[bool] = None,
        include_in_schema: bool = True,
    ) -> None:
        self._prefix = prefix
        self._tags = tags or []
        self._default_response_class = default_response_class
        self.route_class = route_class
        self._deprecated = deprecated
        self._include_in_schema = include_in_schema
        self._responses = responses or {}
        self._routes: List[Union[APIRoute, WebSocketRoute]] = routes or []

    def add_api_route(
        self,
        path: str,
        endpoint: Callable[..., Any],
        *,
        response_model: Optional[Type[Any]] = None,
        status_code: Optional[int] = None,
        tags: Optional[List[str]] = None,
        summary: Optional[str] = None,
        description: Optional[str] = None,
        response_description: str = "Successful Response",
        responses: Optional[Dict[Union[int, str], Dict[str, Any]]] = None,
        deprecated: Optional[bool] = None,
        methods: Optional[Union[Set[str], List[str]]] = None,
        operation_id: Optional[str] = None,
        include_in_schema: bool = True,
        response_class: Type[Response] = JSONResponse,
        name: Optional[str] = None,
        route_class_override: Optional[Type[APIRoute]] = None,
        openapi_extra: Optional[Dict[str, Any]] = None,
    ) -> None:
        route_class = route_class_override or self.route_class
        responses = responses or {}
        combined_responses = {**self._responses, **responses}
        current_response_class = get_value_or_default(
            response_class, self._default_response_class
        )
        current_tags = self._tags.copy()
        if tags:
            current_tags.extend(tags)

        route = route_class(
            self._prefix + path,
            endpoint=endpoint,
            response_model=response_model,
            status_code=status_code,
            tags=current_tags,
            summary=summary,
            description=description,
            response_description=response_description,
            responses=combined_responses,
            deprecated=deprecated or self._deprecated,
            methods=methods,
            operation_id=operation_id,
            include_in_schema=include_in_schema and self._include_in_schema,
            response_class=current_response_class,
            name=name,
            openapi_extra=openapi_extra,
        )
        self.route_register(route)

    def add_route(
        self,
        path: str,
        endpoint: Callable[..., Any],
        *,
        response_model: Optional[Type[Any]] = None,
        status_code: Optional[int] = None,
        tags: Optional[List[str]] = None,
        summary: Optional[str] = None,
        description: Optional[str] = None,
        response_description: str = "Successful Response",
        responses: Optional[Dict[Union[int, str], Dict[str, Any]]] = None,
        deprecated: Optional[bool] = None,
        methods: Optional[Union[Set[str], List[str]]] = None,
        operation_id: Optional[str] = None,
        include_in_schema: bool = True,
        response_class: Type[Response] = Response,
        name: Optional[str] = None,
        route_class_override: Optional[Type[APIRoute]] = None,
        openapi_extra: Optional[Dict[str, Any]] = None,
    ) -> None:
        route_class = route_class_override or self.route_class
        responses = responses or {}
        combined_responses = {**self._responses, **responses}
        current_response_class = get_value_or_default(
            response_class, self._default_response_class
        )
        current_tags = self._tags.copy()
        if tags:
            current_tags.extend(tags)

        route = route_class(
            self._prefix + path,
            endpoint=endpoint,
            response_model=response_model,
            status_code=status_code,
            tags=current_tags,
            summary=summary,
            description=description,
            response_description=response_description,
            responses=combined_responses,
            deprecated=deprecated or self._deprecated,
            methods=methods,
            operation_id=operation_id,
            include_in_schema=include_in_schema and self._include_in_schema,
            response_class=current_response_class,
            name=name,
            openapi_extra=openapi_extra,
        )
        self.route_register(route)

    def route_register(self, route: Union[APIRoute, WebSocketRoute]) -> None:
        self._routes.append(route)

    def add_api(
        self,
        path: str,
        *,
        response_model: Optional[Type[Any]] = None,
        status_code: Optional[int] = None,
        tags: Optional[List[str]] = None,
        summary: Optional[str] = None,
        description: Optional[str] = None,
        response_description: str = "Successful Response",
        responses: Optional[Dict[Union[int, str], Dict[str, Any]]] = None,
        deprecated: Optional[bool] = None,
        methods: Optional[List[str]] = None,
        operation_id: Optional[str] = None,
        include_in_schema: bool = True,
        response_class: Type[Response] = Default(JSONResponse),
        name: Optional[str] = None,
        openapi_extra: Optional[Dict[str, Any]] = None,
    ) -> Callable[[DecoratedCallable], DecoratedCallable]:
        """Registrates API endpoint.

        Examples:
            >>> app = Squall()
            >>>
            >>> @app.add_api("/users/{username}", methods=["GET"])
            >>> async def read_user(username: str):
            >>>     return {"username": username}
        """

        def decorator(func: DecoratedCallable) -> DecoratedCallable:
            self.add_api_route(
                path,
                func,
                response_model=response_model,
                status_code=status_code,
                tags=tags,
                summary=summary,
                description=description,
                response_description=response_description,
                responses=responses,
                deprecated=deprecated,
                methods=methods,
                operation_id=operation_id,
                include_in_schema=include_in_schema,
                response_class=response_class,
                name=name,
                openapi_extra=openapi_extra,
            )
            return func

        return decorator

    def add_ws_route(
        self,
        path: str,
        endpoint: Callable[..., Any],
        *,
        name: Optional[str] = None,
    ) -> None:
        route = WebSocketRoute(
            self._prefix + path,
            endpoint=endpoint,
            name=name,
        )
        self.route_register(route)

    def websocket(
        self, path: str, name: Optional[str] = None
    ) -> Callable[[DecoratedCallable], DecoratedCallable]:
        """Registrates WebSocket endpoint.

        Examples:
            >>> app = Squall()
            >>>
            >>> @app.websocket("/ws")
            >>> async def websocket(websocket: WebSocket):
            >>>     await websocket.accept()
            >>>     await websocket.send_json({"msg": "Hello WebSocket"})
            >>>     await websocket.close()
        """

        def decorator(func: DecoratedCallable) -> DecoratedCallable:
            self.add_ws_route(path, func, name=name)
            return func

        return decorator

    def include_router(self, router: "Router") -> None:
        """Includes child router

        Examples:
            >>> from squall import Squall, Router
            >>> app = Squall()
            >>> router = Router(prefix="/api")
            >>>
            >>> @router.get("/users/{username}")
            >>> async def read_user(username: str):
            >>>     return {"username": username}
            >>>
            >>> app.include_router(router)
        """
        for route in router.routes:
            route.add_path_prefix(self._prefix)
            self.route_register(route)

    @property
    def routes(self) -> List[Union[APIRoute, WebSocketRoute]]:
        return self._routes

    def get(
        self,
        path: str,
        *,
        response_model: Optional[Type[Any]] = None,
        status_code: Optional[int] = None,
        tags: Optional[List[str]] = None,
        summary: Optional[str] = None,
        description: Optional[str] = None,
        response_description: str = "Successful Response",
        responses: Optional[Dict[Union[int, str], Dict[str, Any]]] = None,
        deprecated: Optional[bool] = None,
        operation_id: Optional[str] = None,
        include_in_schema: bool = True,
        response_class: Type[Response] = Default(JSONResponse),
        name: Optional[str] = None,
        openapi_extra: Optional[Dict[str, Any]] = None,
    ) -> Callable[[DecoratedCallable], DecoratedCallable]:
        return self.add_api(
            path=path,
            response_model=response_model,
            status_code=status_code,
            tags=tags,
            summary=summary,
            description=description,
            response_description=response_description,
            responses=responses,
            deprecated=deprecated,
            methods=["GET"],
            operation_id=operation_id,
            include_in_schema=include_in_schema,
            response_class=response_class,
            name=name,
            openapi_extra=openapi_extra,
        )

    def put(
        self,
        path: str,
        *,
        response_model: Optional[Type[Any]] = None,
        status_code: Optional[int] = None,
        tags: Optional[List[str]] = None,
        summary: Optional[str] = None,
        description: Optional[str] = None,
        response_description: str = "Successful Response",
        responses: Optional[Dict[Union[int, str], Dict[str, Any]]] = None,
        deprecated: Optional[bool] = None,
        operation_id: Optional[str] = None,
        include_in_schema: bool = True,
        response_class: Type[Response] = Default(JSONResponse),
        name: Optional[str] = None,
        openapi_extra: Optional[Dict[str, Any]] = None,
    ) -> Callable[[DecoratedCallable], DecoratedCallable]:
        return self.add_api(
            path=path,
            response_model=response_model,
            status_code=status_code,
            tags=tags,
            summary=summary,
            description=description,
            response_description=response_description,
            responses=responses,
            deprecated=deprecated,
            methods=["PUT"],
            operation_id=operation_id,
            include_in_schema=include_in_schema,
            response_class=response_class,
            name=name,
            openapi_extra=openapi_extra,
        )

    def post(
        self,
        path: str,
        *,
        response_model: Optional[Type[Any]] = None,
        status_code: Optional[int] = None,
        tags: Optional[List[str]] = None,
        summary: Optional[str] = None,
        description: Optional[str] = None,
        response_description: str = "Successful Response",
        responses: Optional[Dict[Union[int, str], Dict[str, Any]]] = None,
        deprecated: Optional[bool] = None,
        operation_id: Optional[str] = None,
        include_in_schema: bool = True,
        response_class: Type[Response] = Default(JSONResponse),
        name: Optional[str] = None,
        openapi_extra: Optional[Dict[str, Any]] = None,
    ) -> Callable[[DecoratedCallable], DecoratedCallable]:
        return self.add_api(
            path=path,
            response_model=response_model,
            status_code=status_code,
            tags=tags,
            summary=summary,
            description=description,
            response_description=response_description,
            responses=responses,
            deprecated=deprecated,
            methods=["POST"],
            operation_id=operation_id,
            include_in_schema=include_in_schema,
            response_class=response_class,
            name=name,
            openapi_extra=openapi_extra,
        )

    def delete(
        self,
        path: str,
        *,
        response_model: Optional[Type[Any]] = None,
        status_code: Optional[int] = None,
        tags: Optional[List[str]] = None,
        summary: Optional[str] = None,
        description: Optional[str] = None,
        response_description: str = "Successful Response",
        responses: Optional[Dict[Union[int, str], Dict[str, Any]]] = None,
        deprecated: Optional[bool] = None,
        operation_id: Optional[str] = None,
        include_in_schema: bool = True,
        response_class: Type[Response] = Default(JSONResponse),
        name: Optional[str] = None,
        openapi_extra: Optional[Dict[str, Any]] = None,
    ) -> Callable[[DecoratedCallable], DecoratedCallable]:
        return self.add_api(
            path=path,
            response_model=response_model,
            status_code=status_code,
            tags=tags,
            summary=summary,
            description=description,
            response_description=response_description,
            responses=responses,
            deprecated=deprecated,
            methods=["DELETE"],
            operation_id=operation_id,
            include_in_schema=include_in_schema,
            response_class=response_class,
            name=name,
            openapi_extra=openapi_extra,
        )

    def options(
        self,
        path: str,
        *,
        response_model: Optional[Type[Any]] = None,
        status_code: Optional[int] = None,
        tags: Optional[List[str]] = None,
        summary: Optional[str] = None,
        description: Optional[str] = None,
        response_description: str = "Successful Response",
        responses: Optional[Dict[Union[int, str], Dict[str, Any]]] = None,
        deprecated: Optional[bool] = None,
        operation_id: Optional[str] = None,
        include_in_schema: bool = True,
        response_class: Type[Response] = Default(JSONResponse),
        name: Optional[str] = None,
        openapi_extra: Optional[Dict[str, Any]] = None,
    ) -> Callable[[DecoratedCallable], DecoratedCallable]:
        return self.add_api(
            path=path,
            response_model=response_model,
            status_code=status_code,
            tags=tags,
            summary=summary,
            description=description,
            response_description=response_description,
            responses=responses,
            deprecated=deprecated,
            methods=["OPTIONS"],
            operation_id=operation_id,
            include_in_schema=include_in_schema,
            response_class=response_class,
            name=name,
            openapi_extra=openapi_extra,
        )

    def head(
        self,
        path: str,
        *,
        response_model: Optional[Type[Any]] = None,
        status_code: Optional[int] = None,
        tags: Optional[List[str]] = None,
        summary: Optional[str] = None,
        description: Optional[str] = None,
        response_description: str = "Successful Response",
        responses: Optional[Dict[Union[int, str], Dict[str, Any]]] = None,
        deprecated: Optional[bool] = None,
        operation_id: Optional[str] = None,
        include_in_schema: bool = True,
        response_class: Type[Response] = Default(JSONResponse),
        name: Optional[str] = None,
        openapi_extra: Optional[Dict[str, Any]] = None,
    ) -> Callable[[DecoratedCallable], DecoratedCallable]:
        return self.add_api(
            path=path,
            response_model=response_model,
            status_code=status_code,
            tags=tags,
            summary=summary,
            description=description,
            response_description=response_description,
            responses=responses,
            deprecated=deprecated,
            methods=["HEAD"],
            operation_id=operation_id,
            include_in_schema=include_in_schema,
            response_class=response_class,
            name=name,
            openapi_extra=openapi_extra,
        )

    def patch(
        self,
        path: str,
        *,
        response_model: Optional[Type[Any]] = None,
        status_code: Optional[int] = None,
        tags: Optional[List[str]] = None,
        summary: Optional[str] = None,
        description: Optional[str] = None,
        response_description: str = "Successful Response",
        responses: Optional[Dict[Union[int, str], Dict[str, Any]]] = None,
        deprecated: Optional[bool] = None,
        operation_id: Optional[str] = None,
        include_in_schema: bool = True,
        response_class: Type[Response] = Default(JSONResponse),
        name: Optional[str] = None,
        openapi_extra: Optional[Dict[str, Any]] = None,
    ) -> Callable[[DecoratedCallable], DecoratedCallable]:
        return self.add_api(
            path=path,
            response_model=response_model,
            status_code=status_code,
            tags=tags,
            summary=summary,
            description=description,
            response_description=response_description,
            responses=responses,
            deprecated=deprecated,
            methods=["PATCH"],
            operation_id=operation_id,
            include_in_schema=include_in_schema,
            response_class=response_class,
            name=name,
            openapi_extra=openapi_extra,
        )

    def trace(
        self,
        path: str,
        *,
        response_model: Optional[Type[Any]] = None,
        status_code: Optional[int] = None,
        tags: Optional[List[str]] = None,
        summary: Optional[str] = None,
        description: Optional[str] = None,
        response_description: str = "Successful Response",
        responses: Optional[Dict[Union[int, str], Dict[str, Any]]] = None,
        deprecated: Optional[bool] = None,
        operation_id: Optional[str] = None,
        include_in_schema: bool = True,
        response_class: Type[Response] = Default(JSONResponse),
        name: Optional[str] = None,
        openapi_extra: Optional[Dict[str, Any]] = None,
    ) -> Callable[[DecoratedCallable], DecoratedCallable]:

        return self.add_api(
            path=path,
            response_model=response_model,
            status_code=status_code,
            tags=tags,
            summary=summary,
            description=description,
            response_description=response_description,
            responses=responses,
            deprecated=deprecated,
            methods=["TRACE"],
            operation_id=operation_id,
            include_in_schema=include_in_schema,
            response_class=response_class,
            name=name,
            openapi_extra=openapi_extra,
        )


class RootRouter(Router):
    def __init__(
        self,
        prefix: str = "",
        tags: Optional[List[str]] = None,
        default_response_class: Type[Response] = Default(JSONResponse),
        responses: Optional[Dict[Union[int, str], Dict[str, Any]]] = None,
        routes: Optional[List[Union[APIRoute, WebSocketRoute]]] = None,
        redirect_slashes: bool = True,
        default: Optional[ASGIApp] = None,
        route_class: Type[APIRoute] = APIRoute,
        deprecated: Optional[bool] = None,
        include_in_schema: bool = True,
    ) -> None:
        # Need both, Router and Router
        super(RootRouter, self).__init__(
            prefix=prefix,
            tags=tags,
            default_response_class=default_response_class,
            routes=routes,
            route_class=route_class,
            deprecated=deprecated,
            include_in_schema=include_in_schema,
            responses=responses,
        )
        self.redirect_slashes = redirect_slashes
        self.default = default or self.not_found
        self._fast_path_route_http: Dict[str, Any] = {}
        self._fast_path_route_ws: Dict[str, Any] = {}

    def __call__(self, scope: Scope, receive: Receive, send: Send) -> Awaitable[Any]:
        """
        The main entry point to the Router class.
        """
        if "router" not in scope:
            scope["router"] = self

        routes: Sequence[Union[APIRoute, WebSocketRoute]]
        if scope["type"] == "http":
            routes, _locations = self.get_http_routes(scope["method"], scope["path"])
        elif scope["type"] == "websocket":
            routes, _ = self.get_ws_routes(scope["path"]), []
        else:
            assert False, f"RootRouter doesn't allow scope type: {scope['type']}"

        for route in routes:
            # Determine if any route matches the incoming scope,
            # and hand over to the matching route if found.
            match, child_scope = route.matches(scope)
            if match:
                scope.update(child_scope)
                return route.app(scope, receive, send)

        # for location in reversed(_locations):
        #     match, child_scope = location.matches(scope)
        #     if match:
        #         scope.update(child_scope)
        #         await location.handle(scope, receive, send)
        #         return

        if scope["type"] == "http" and self.redirect_slashes and scope["path"] != "/":
            redirect_scope = dict(scope)
            if scope["path"].endswith("/"):
                redirect_scope["path"] = redirect_scope["path"].rstrip("/")
            else:
                redirect_scope["path"] = redirect_scope["path"] + "/"

            routes, _ = self.get_http_routes(
                redirect_scope["method"], redirect_scope["path"]
            )
            for route in routes:
                match, child_scope = route.matches(redirect_scope)
                if match:
                    redirect_url = URL(scope=redirect_scope)
                    response = RedirectResponse(url=str(redirect_url))
                    return response(scope, receive, send)

        return self.default(scope, receive, send)

    @staticmethod
    async def not_found(scope: Scope, receive: Receive, send: Send) -> None:
        if scope["type"] == "websocket":
            websocket_close = WebSocketClose()
            await websocket_close(scope, receive, send)
            return

        # If we're running inside a starlette application then raise an
        # exception, so that the configurable exception handler can deal with
        # returning the response. For plain ASGI apps, just return the response.
        if "app" in scope:
            raise HTTPException(status_code=404)
        else:
            response = PlainTextResponse("Not Found", status_code=404)
        await response(scope, receive, send)

    @staticmethod
    def _get_fast_path_octets(path: str) -> List[str]:
        no_regex = re.sub(r"(\([^)]*\))", PLACEHOLDER_DYNAMIC, path)
        no_format = re.sub(r"{([^}]*)}", PLACEHOLDER_DYNAMIC, no_regex)
        result = []
        for i in no_format.strip("/").split("/"):
            value = PLACEHOLDER_DYNAMIC if PLACEHOLDER_DYNAMIC in i else i
            result.append(value)
        return result

    def _add_fast_path_route(
        self, route: Union[APIRoute, WebSocketRoute], websocket: bool = False
    ) -> None:
        layer = self._fast_path_route_ws if websocket else self._fast_path_route_http

        for octet in self._get_fast_path_octets(route.path):
            if octet not in layer:
                layer[octet] = {}
                if octet != PLACEHOLDER_DYNAMIC:
                    layer[octet][PLACEHOLDER_DYNAMIC] = {}
            layer = layer[octet]

        if "#routes#" not in layer:
            layer["#routes#"] = [] if websocket else {}

        if websocket:
            layer["#routes#"].append(route)
            return

        # List here for handling cases when single
        # octets path can have different patterns
        # /some/{number_item:int}
        # /some/{string_item:str}
        methods = getattr(route, "methods", [])
        for method in methods:
            if method not in layer["#routes#"]:
                layer["#routes#"][method] = []
            layer["#routes#"][method].append(route)

    def route_register(self, route: Union[APIRoute, WebSocketRoute]) -> None:
        if isinstance(route, APIRoute):
            self._add_fast_path_route(route, websocket=False)
        elif isinstance(route, WebSocketRoute):
            self._add_fast_path_route(route, websocket=True)
        self._routes.append(route)

    def get_http_routes(
        self, method: str, path: str
    ) -> Tuple[List[APIRoute], List[APIRoute]]:
        last = self._fast_path_route_http
        locations = []
        get = dict.get
        try:
            for key in path.strip("/").split("/"):
                last = get(last, key) or last[PLACEHOLDER_DYNAMIC]
                if PLACEHOLDER_LOCATION in last:
                    locations.append(last[PLACEHOLDER_LOCATION])
        except KeyError:
            return [], locations
        routes: List[APIRoute] = get(get(last, "#routes#", {}), method, [])
        return routes, locations

    def get_ws_routes(self, path: str) -> List[WebSocketRoute]:
        last = self._fast_path_route_ws
        get = dict.get
        try:
            for key in path.strip("/").split("/"):
                last = get(last, key) or last[PLACEHOLDER_DYNAMIC]
        except KeyError:
            return []
        return last.get("#routes#", [])



import pytest
from squall import UploadFile
from squall.datastructures import Default


def test_upload_file_invalid():
    with pytest.raises(ValueError):
        UploadFile.validate("not a Starlette UploadFile")


def test_default_placeholder_equals():
    placeholder_1 = Default("a")
    placeholder_2 = Default("a")
    assert placeholder_1 == placeholder_2
    assert placeholder_1.value == placeholder_2.value


def test_default_placeholder_bool():
    placeholder_a = Default("a")
    placeholder_b = Default("")
    assert placeholder_a
    assert not placeholder_b



from dataclasses import dataclass

from squall import Body, Cookie, Header, Path, Query, Squall
from squall.testclient import TestClient

app = Squall()


@dataclass
class Item:
    data: str


@app.post("/schema_extra/")
def schema_extra(item: Item):
    return item


@app.post("/example/")
def example(item: Item = Body(..., example={"data": "Data in Body example"})):
    return item


@app.post("/examples/")
def examples(
    item: Item = Body(
        ...,
        examples={
            "example1": {
                "summary": "example1 summary",
                "value": {"data": "Data in Body examples, example1"},
            },
            "example2": {"value": {"data": "Data in Body examples, example2"}},
        },
    )
):
    return item


@app.post("/example_examples/")
def example_examples(
    item: Item = Body(
        ...,
        example={"data": "Overriden example"},
        examples={
            "example1": {"value": {"data": "examples example_examples 1"}},
            "example2": {"value": {"data": "examples example_examples 2"}},
        },
    )
):
    return item


# TODO: enable these tests once/if Form(embed=False) is supported
# TODO: In that case, define if File() should support example/examples too
# @app.post("/form_example")
# def form_example(firstname: str = Form(..., example="John")):
#     return firstname


# @app.post("/form_examples")
# def form_examples(
#     lastname: str = Form(
#         ...,
#         examples={
#             "example1": {"summary": "last name summary", "value": "Doe"},
#             "example2": {"value": "Doesn't"},
#         },
#     ),
# ):
#     return lastname


# @app.post("/form_example_examples")
# def form_example_examples(
#     lastname: str = Form(
#         ...,
#         example="Doe overriden",
#         examples={
#             "example1": {"summary": "last name summary", "value": "Doe"},
#             "example2": {"value": "Doesn't"},
#         },
#     ),
# ):
#     return lastname


@app.get("/path_example/{item_id}")
def path_example(
    item_id: str = Path(
        ...,
        example="item_1",
    ),
):
    return item_id


@app.get("/path_examples/{item_id}")
def path_examples(
    item_id: str = Path(
        ...,
        examples={
            "example1": {"summary": "item ID summary", "value": "item_1"},
            "example2": {"value": "item_2"},
        },
    ),
):
    return item_id


@app.get("/path_example_examples/{item_id}")
def path_example_examples(
    item_id: str = Path(
        ...,
        example="item_overriden",
        examples={
            "example1": {"summary": "item ID summary", "value": "item_1"},
            "example2": {"value": "item_2"},
        },
    ),
):
    return item_id


@app.get("/query_example/")
def query_example(
    data: str = Query(
        None,
        example="query1",
    ),
):
    return data


@app.get("/query_examples/")
def query_examples(
    data: str = Query(
        None,
        examples={
            "example1": {"summary": "Query example 1", "value": "query1"},
            "example2": {"value": "query2"},
        },
    ),
):
    return data


@app.get("/query_example_examples/")
def query_example_examples(
    data: str = Query(
        None,
        example="query_overriden",
        examples={
            "example1": {"summary": "Qeury example 1", "value": "query1"},
            "example2": {"value": "query2"},
        },
    ),
):
    return data


@app.get("/header_example/")
def header_example(
    data: str = Header(
        None,
        example="header1",
    ),
):
    return data


@app.get("/header_examples/")
def header_examples(
    data: str = Header(
        None,
        examples={
            "example1": {"summary": "header example 1", "value": "header1"},
            "example2": {"value": "header2"},
        },
    ),
):
    return data


@app.get("/header_example_examples/")
def header_example_examples(
    data: str = Header(
        None,
        example="header_overriden",
        examples={
            "example1": {"summary": "Qeury example 1", "value": "header1"},
            "example2": {"value": "header2"},
        },
    ),
):
    return data


@app.get("/cookie_example/")
def cookie_example(
    data: str = Cookie(
        None,
        example="cookie1",
    ),
):
    return data


@app.get("/cookie_examples/")
def cookie_examples(
    data: str = Cookie(
        None,
        examples={
            "example1": {"summary": "cookie example 1", "value": "cookie1"},
            "example2": {"value": "cookie2"},
        },
    ),
):
    return data


@app.get("/cookie_example_examples/")
def cookie_example_examples(
    data: str = Cookie(
        None,
        example="cookie_overriden",
        examples={
            "example1": {"summary": "Qeury example 1", "value": "cookie1"},
            "example2": {"value": "cookie2"},
        },
    ),
):
    return data


client = TestClient(app)


openapi_schema = {
    "openapi": "3.0.2",
    "info": {"title": "Squall", "version": "0.1.0"},
    "components": {
        "schemas": {
            "Item": {
                "type": "object",
                "properties": {"data": {"type": "string"}},
                "required": ["data"],
                "additionalProperties": False,
            },
            "ValidationError": {
                "title": "ValidationError",
                "type": "object",
                "properties": {
                    "loc": {
                        "title": "Location",
                        "type": "array",
                        "items": {"type": "string"},
                    },
                    "msg": {"title": "Message", "type": "string"},
                    "type": {"title": "Error Type", "type": "string"},
                },
                "required": ["loc", "msg", "type"],
            },
            "HTTPValidationError": {
                "title": "HTTPValidationError",
                "type": "object",
                "properties": {
                    "detail": {
                        "title": "Detail",
                        "type": "array",
                        "items": {"$ref": "#/components/schemas/ValidationError"},
                    }
                },
            },
            "HTTPBadRequestError": {
                "title": "HTTPBadRequestError",
                "type": "object",
                "properties": {
                    "details": {
                        "title": "Detail",
                        "type": "array",
                        "items": {"$ref": "#/components/schemas/ValidationError"},
                    }
                },
            },
        }
    },
    "paths": {
        "/schema_extra/": {
            "post": {
                "summary": "Schema Extra",
                "operationId": "schema_extra_schema_extra__post",
                "responses": {
                    "200": {
                        "description": "Successful Response",
                        "content": {"application/json": {"schema": {}}},
                    },
                    "422": {
                        "description": "Request Body Validation Error",
                        "content": {
                            "application/json": {
                                "schema": {
                                    "$ref": "#/components/schemas/HTTPValidationError"
                                }
                            }
                        },
                    },
                },
                "requestBody": {
                    "required": True,
                    "content": {
                        "application/json": {
                            "schema": {"$ref": "#/components/schemas/Item"}
                        }
                    },
                },
            }
        },
        "/example/": {
            "post": {
                "summary": "Example",
                "operationId": "example_example__post",
                "responses": {
                    "200": {
                        "description": "Successful Response",
                        "content": {"application/json": {"schema": {}}},
                    },
                    "422": {
                        "description": "Request Body Validation Error",
                        "content": {
                            "application/json": {
                                "schema": {
                                    "$ref": "#/components/schemas/HTTPValidationError"
                                }
                            }
                        },
                    },
                },
                "requestBody": {
                    "required": None,
                    "content": {
                        "application/json": {
                            "example": {"data": "Data in Body example"},
                            "schema": {"$ref": "#/components/schemas/Item"},
                        }
                    },
                },
            }
        },
        "/examples/": {
            "post": {
                "summary": "Examples",
                "operationId": "examples_examples__post",
                "responses": {
                    "200": {
                        "description": "Successful Response",
                        "content": {"application/json": {"schema": {}}},
                    },
                    "422": {
                        "description": "Request Body Validation Error",
                        "content": {
                            "application/json": {
                                "schema": {
                                    "$ref": "#/components/schemas/HTTPValidationError"
                                }
                            }
                        },
                    },
                },
                "requestBody": {
                    "required": None,
                    "content": {
                        "application/json": {
                            "examples": {
                                "example1": {
                                    "summary": "example1 summary",
                                    "value": {
                                        "data": "Data in Body examples, example1"
                                    },
                                },
                                "example2": {
                                    "value": {"data": "Data in Body examples, example2"}
                                },
                            },
                            "schema": {"$ref": "#/components/schemas/Item"},
                        }
                    },
                },
            }
        },
        "/example_examples/": {
            "post": {
                "summary": "Example Examples",
                "operationId": "example_examples_example_examples__post",
                "responses": {
                    "200": {
                        "description": "Successful Response",
                        "content": {"application/json": {"schema": {}}},
                    },
                    "422": {
                        "description": "Request Body Validation Error",
                        "content": {
                            "application/json": {
                                "schema": {
                                    "$ref": "#/components/schemas/HTTPValidationError"
                                }
                            }
                        },
                    },
                },
                "requestBody": {
                    "required": None,
                    "content": {
                        "application/json": {
                            "examples": {
                                "example1": {
                                    "value": {"data": "examples example_examples 1"}
                                },
                                "example2": {
                                    "value": {"data": "examples example_examples 2"}
                                },
                            },
                            "schema": {"$ref": "#/components/schemas/Item"},
                        }
                    },
                },
            }
        },
        "/path_example/{item_id}": {
            "get": {
                "summary": "Path Example",
                "operationId": "path_example_path_example__item_id__get",
                "parameters": [
                    {
                        "required": True,
                        "schema": {"type": "string"},
                        "name": "item_id",
                        "in": "path",
                        "example": "item_1",
                    }
                ],
                "responses": {
                    "200": {
                        "description": "Successful Response",
                        "content": {"application/json": {"schema": {}}},
                    },
                    "400": {
                        "description": "Parameters Validation Error",
                        "content": {
                            "application/json": {
                                "schema": {
                                    "$ref": "#/components/schemas/HTTPBadRequestError"
                                }
                            }
                        },
                    },
                },
            }
        },
        "/path_examples/{item_id}": {
            "get": {
                "summary": "Path Examples",
                "operationId": "path_examples_path_examples__item_id__get",
                "parameters": [
                    {
                        "required": True,
                        "schema": {"type": "string"},
                        "name": "item_id",
                        "in": "path",
                        "examples": {
                            "example1": {
                                "summary": "item ID summary",
                                "value": "item_1",
                            },
                            "example2": {"value": "item_2"},
                        },
                    }
                ],
                "responses": {
                    "200": {
                        "description": "Successful Response",
                        "content": {"application/json": {"schema": {}}},
                    },
                    "400": {
                        "description": "Parameters Validation Error",
                        "content": {
                            "application/json": {
                                "schema": {
                                    "$ref": "#/components/schemas/HTTPBadRequestError"
                                }
                            }
                        },
                    },
                },
            }
        },
        "/path_example_examples/{item_id}": {
            "get": {
                "summary": "Path Example Examples",
                "operationId": "path_example_examples_path_example_examples__item_id__get",
                "parameters": [
                    {
                        "required": True,
                        "schema": {"type": "string"},
                        "name": "item_id",
                        "in": "path",
                        "example": "item_overriden",
                        "examples": {
                            "example1": {
                                "summary": "item ID summary",
                                "value": "item_1",
                            },
                            "example2": {"value": "item_2"},
                        },
                    }
                ],
                "responses": {
                    "200": {
                        "description": "Successful Response",
                        "content": {"application/json": {"schema": {}}},
                    },
                    "400": {
                        "description": "Parameters Validation Error",
                        "content": {
                            "application/json": {
                                "schema": {
                                    "$ref": "#/components/schemas/HTTPBadRequestError"
                                }
                            }
                        },
                    },
                },
            }
        },
        "/query_example/": {
            "get": {
                "summary": "Query Example",
                "operationId": "query_example_query_example__get",
                "parameters": [
                    {
                        "required": False,
                        "schema": {"type": "string"},
                        "name": "data",
                        "in": "query",
                        "example": "query1",
                    }
                ],
                "responses": {
                    "200": {
                        "description": "Successful Response",
                        "content": {"application/json": {"schema": {}}},
                    },
                    "400": {
                        "description": "Parameters Validation Error",
                        "content": {
                            "application/json": {
                                "schema": {
                                    "$ref": "#/components/schemas/HTTPBadRequestError"
                                }
                            }
                        },
                    },
                },
            }
        },
        "/query_examples/": {
            "get": {
                "summary": "Query Examples",
                "operationId": "query_examples_query_examples__get",
                "parameters": [
                    {
                        "required": False,
                        "schema": {"type": "string"},
                        "name": "data",
                        "in": "query",
                        "examples": {
                            "example1": {
                                "summary": "Query example 1",
                                "value": "query1",
                            },
                            "example2": {"value": "query2"},
                        },
                    }
                ],
                "responses": {
                    "200": {
                        "description": "Successful Response",
                        "content": {"application/json": {"schema": {}}},
                    },
                    "400": {
                        "description": "Parameters Validation Error",
                        "content": {
                            "application/json": {
                                "schema": {
                                    "$ref": "#/components/schemas/HTTPBadRequestError"
                                }
                            }
                        },
                    },
                },
            }
        },
        "/query_example_examples/": {
            "get": {
                "summary": "Query Example Examples",
                "operationId": "query_example_examples_query_example_examples__get",
                "parameters": [
                    {
                        "required": False,
                        "schema": {"type": "string"},
                        "name": "data",
                        "in": "query",
                        "example": "query_overriden",
                        "examples": {
                            "example1": {
                                "summary": "Qeury example 1",
                                "value": "query1",
                            },
                            "example2": {"value": "query2"},
                        },
                    }
                ],
                "responses": {
                    "200": {
                        "description": "Successful Response",
                        "content": {"application/json": {"schema": {}}},
                    },
                    "400": {
                        "description": "Parameters Validation Error",
                        "content": {
                            "application/json": {
                                "schema": {
                                    "$ref": "#/components/schemas/HTTPBadRequestError"
                                }
                            }
                        },
                    },
                },
            }
        },
        "/header_example/": {
            "get": {
                "summary": "Header Example",
                "operationId": "header_example_header_example__get",
                "parameters": [
                    {
                        "required": False,
                        "schema": {"type": "string"},
                        "name": "data",
                        "in": "header",
                        "example": "header1",
                    }
                ],
                "responses": {
                    "200": {
                        "description": "Successful Response",
                        "content": {"application/json": {"schema": {}}},
                    },
                    "400": {
                        "description": "Parameters Validation Error",
                        "content": {
                            "application/json": {
                                "schema": {
                                    "$ref": "#/components/schemas/HTTPBadRequestError"
                                }
                            }
                        },
                    },
                },
            }
        },
        "/header_examples/": {
            "get": {
                "summary": "Header Examples",
                "operationId": "header_examples_header_examples__get",
                "parameters": [
                    {
                        "required": False,
                        "schema": {"type": "string"},
                        "name": "data",
                        "in": "header",
                        "examples": {
                            "example1": {
                                "summary": "header example 1",
                                "value": "header1",
                            },
                            "example2": {"value": "header2"},
                        },
                    }
                ],
                "responses": {
                    "200": {
                        "description": "Successful Response",
                        "content": {"application/json": {"schema": {}}},
                    },
                    "400": {
                        "description": "Parameters Validation Error",
                        "content": {
                            "application/json": {
                                "schema": {
                                    "$ref": "#/components/schemas/HTTPBadRequestError"
                                }
                            }
                        },
                    },
                },
            }
        },
        "/header_example_examples/": {
            "get": {
                "summary": "Header Example Examples",
                "operationId": "header_example_examples_header_example_examples__get",
                "parameters": [
                    {
                        "required": False,
                        "schema": {"type": "string"},
                        "name": "data",
                        "in": "header",
                        "example": "header_overriden",
                        "examples": {
                            "example1": {
                                "summary": "Qeury example 1",
                                "value": "header1",
                            },
                            "example2": {"value": "header2"},
                        },
                    }
                ],
                "responses": {
                    "200": {
                        "description": "Successful Response",
                        "content": {"application/json": {"schema": {}}},
                    },
                    "400": {
                        "description": "Parameters Validation Error",
                        "content": {
                            "application/json": {
                                "schema": {
                                    "$ref": "#/components/schemas/HTTPBadRequestError"
                                }
                            }
                        },
                    },
                },
            }
        },
        "/cookie_example/": {
            "get": {
                "summary": "Cookie Example",
                "operationId": "cookie_example_cookie_example__get",
                "parameters": [
                    {
                        "required": False,
                        "schema": {"type": "string"},
                        "name": "data",
                        "in": "cookie",
                        "example": "cookie1",
                    }
                ],
                "responses": {
                    "200": {
                        "description": "Successful Response",
                        "content": {"application/json": {"schema": {}}},
                    },
                    "400": {
                        "description": "Parameters Validation Error",
                        "content": {
                            "application/json": {
                                "schema": {
                                    "$ref": "#/components/schemas/HTTPBadRequestError"
                                }
                            }
                        },
                    },
                },
            }
        },
        "/cookie_examples/": {
            "get": {
                "summary": "Cookie Examples",
                "operationId": "cookie_examples_cookie_examples__get",
                "parameters": [
                    {
                        "required": False,
                        "schema": {"type": "string"},
                        "name": "data",
                        "in": "cookie",
                        "examples": {
                            "example1": {
                                "summary": "cookie example 1",
                                "value": "cookie1",
                            },
                            "example2": {"value": "cookie2"},
                        },
                    }
                ],
                "responses": {
                    "200": {
                        "description": "Successful Response",
                        "content": {"application/json": {"schema": {}}},
                    },
                    "400": {
                        "description": "Parameters Validation Error",
                        "content": {
                            "application/json": {
                                "schema": {
                                    "$ref": "#/components/schemas/HTTPBadRequestError"
                                }
                            }
                        },
                    },
                },
            }
        },
        "/cookie_example_examples/": {
            "get": {
                "summary": "Cookie Example Examples",
                "operationId": "cookie_example_examples_cookie_example_examples__get",
                "parameters": [
                    {
                        "required": False,
                        "schema": {"type": "string"},
                        "name": "data",
                        "in": "cookie",
                        "example": "cookie_overriden",
                        "examples": {
                            "example1": {
                                "summary": "Qeury example 1",
                                "value": "cookie1",
                            },
                            "example2": {"value": "cookie2"},
                        },
                    }
                ],
                "responses": {
                    "200": {
                        "description": "Successful Response",
                        "content": {"application/json": {"schema": {}}},
                    },
                    "400": {
                        "description": "Parameters Validation Error",
                        "content": {
                            "application/json": {
                                "schema": {
                                    "$ref": "#/components/schemas/HTTPBadRequestError"
                                }
                            }
                        },
                    },
                },
            }
        },
    },
}


def test_openapi_schema():
    """
    Test that example overrides work:

    * model schema_extra is included
    * Body(example={}) overrides schema_extra in model
    * Body(examples{}) overrides Body(example={}) and schema_extra in model
    """
    response = client.get("/openapi.json")
    assert response.status_code == 200, response.text
    assert response.json() == openapi_schema


def test_call_api():
    response = client.post("/schema_extra/", json={"data": "Foo"})
    assert response.status_code == 200, response.text
    response = client.post("/example/", json={"data": "Foo"})
    assert response.status_code == 200, response.text
    response = client.post("/examples/", json={"data": "Foo"})
    assert response.status_code == 200, response.text
    response = client.post("/example_examples/", json={"data": "Foo"})
    assert response.status_code == 200, response.text
    response = client.get("/path_example/foo")
    assert response.status_code == 200, response.text
    response = client.get("/path_examples/foo")
    assert response.status_code == 200, response.text
    response = client.get("/path_example_examples/foo")
    assert response.status_code == 200, response.text
    response = client.get("/query_example/")
    assert response.status_code == 200, response.text
    response = client.get("/query_examples/")
    assert response.status_code == 200, response.text
    response = client.get("/query_example_examples/")
    assert response.status_code == 200, response.text
    response = client.get("/header_example/")
    assert response.status_code == 200, response.text
    response = client.get("/header_examples/")
    assert response.status_code == 200, response.text
    response = client.get("/header_example_examples/")
    assert response.status_code == 200, response.text
    response = client.get("/cookie_example/")
    assert response.status_code == 200, response.text
    response = client.get("/cookie_examples/")
    assert response.status_code == 200, response.text
    response = client.get("/cookie_example_examples/")
    assert response.status_code == 200, response.text



from squall import Router, Squall
from squall.testclient import TestClient

router = Router()

sub_router = Router(prefix="/items")

app = Squall()


@sub_router.get("/")
def read_item():
    return {"id": "foo"}


router.include_router(sub_router)

app.include_router(router)


openapi_schema = {
    "openapi": "3.0.2",
    "info": {"title": "Squall", "version": "0.1.0"},
    "paths": {
        "/items/": {
            "get": {
                "responses": {
                    "200": {
                        "description": "Successful Response",
                        "content": {"application/json": {"schema": {}}},
                    }
                },
                "summary": "Read Item",
                "operationId": "read_item_items__get",
            }
        }
    },
}

client = TestClient(app)


def test_openapi_schema():
    response = client.get("/openapi.json")
    assert response.status_code == 200, response.text
    assert response.json() == openapi_schema


def test_path_operation():
    response = client.get("/items/")
    assert response.status_code == 200, response.text
    assert response.json() == {"id": "foo"}



import inspect
import re
from typing import Any, Callable, Dict, Set, Union, get_args, get_origin

from squall.datastructures import DefaultPlaceholder, DefaultType


def get_path_param_names(path: str) -> Set[str]:
    return set(re.findall("{(.*?)}", path))


def generate_operation_id_for_path(*, name: str, path: str, method: str) -> str:
    operation_id = name + path
    operation_id = re.sub("[^0-9a-zA-Z_]", "_", operation_id)
    operation_id = operation_id + "_" + method.lower()
    return operation_id


def get_callable_name(callable: Callable[..., Any]) -> str:
    if inspect.isfunction(callable) or inspect.isclass(callable):
        return callable.__name__
    return callable.__class__.__name__


def deep_dict_update(main_dict: Dict[Any, Any], update_dict: Dict[Any, Any]) -> None:
    for key, value in update_dict.items():
        if (
            key in main_dict
            and isinstance(main_dict[key], dict)
            and isinstance(update_dict[key], dict)
        ):
            deep_dict_update(main_dict[key], value)
        else:
            main_dict[key] = update_dict[key]


def get_value_or_default(
    first_item: Union[DefaultPlaceholder, DefaultType],
    *extra_items: Union[DefaultPlaceholder, DefaultType],
) -> Union[DefaultPlaceholder, DefaultType]:
    """
    Pass items or `DefaultPlaceholder`s by descending priority.

    The first one to _not_ be a `DefaultPlaceholder` will be returned.

    Otherwise, the first item (a `DefaultPlaceholder`) will be returned.
    """
    items = (first_item,) + extra_items
    for item in items:
        if not isinstance(item, DefaultPlaceholder):
            return item
    return first_item


def get_types(annotation: Any) -> Set[Any]:
    """Returns all types in the annotation

    :param annotation: variable to get types from
    :returns: set of available types

    Example:
        >>> get_types(Optional[List[UserInfoResponse]])
        {typing.Union, <class 'NoneType'>, <class '__main__.UserInfoResponse'>, <class 'list'>}
    """
    result = set()

    if inspect.isclass(annotation):
        result.add(annotation)

    if alias := get_origin(annotation):
        result.add(alias)
        result.update(get_types(alias))

    for i in get_args(annotation):
        if not get_origin(i):
            result.add(i)
        else:
            result.update(get_types(i))
    return result



import http
from typing import Optional

import pytest
from squall import Squall
from squall.params import Num, Path, Query, Str
from squall.testclient import TestClient

app = Squall()


@app.add_api("/api_route")
def non_operation():
    return {"message": "Hello World"}


def non_decorated_route():
    return {"message": "Hello World"}


app.add_api_route("/non_decorated_route", non_decorated_route)


@app.get("/text")
def get_text():
    return "Hello World"


@app.get("/path/{item_id}")
def get_id(item_id):
    return item_id


@app.get("/path/str/{item_id}")
def get_str_id(item_id: str):
    return item_id


@app.get("/path/int/{item_id}")
def get_int_id(item_id: int):
    return item_id


@app.get("/path/float/{item_id}")
def get_float_id(item_id: float):
    return item_id


@app.get("/path/param/{item_id}")
def get_path_param_id(item_id: Optional[str] = Path(None)):
    return item_id


@app.get("/path/param-required/{item_id}")
def get_path_param_required_id(item_id: str = Path(...)):
    return item_id


@app.get("/path/param-minlength/{item_id}")
def get_path_param_min_len(item_id: str = Path(..., valid=Str(min_len=3))):
    return item_id


@app.get("/path/param-maxlength/{item_id}")
def get_path_param_max_len(item_id: str = Path(..., valid=Str(max_len=3))):
    return item_id


@app.get("/path/param-min_maxlength/{item_id}")
def get_path_param_min_max_len(
    item_id: str = Path(..., valid=Str(max_len=3, min_len=2))
):
    return item_id


@app.get("/path/param-gt/{item_id}")
def get_path_param_gt(item_id: float = Path(..., valid=Num(gt=3))):
    return item_id


@app.get("/path/param-gt0/{item_id}")
def get_path_param_gt0(item_id: float = Path(..., valid=Num(gt=0))):
    return item_id


@app.get("/path/param-ge/{item_id}")
def get_path_param_ge(item_id: float = Path(..., valid=Num(ge=3))):
    return item_id


@app.get("/path/param-lt/{item_id}")
def get_path_param_lt(item_id: float = Path(..., valid=Num(lt=3))):
    return item_id


@app.get("/path/param-lt0/{item_id}")
def get_path_param_lt0(item_id: float = Path(..., valid=Num(lt=0))):
    return item_id


@app.get("/path/param-le/{item_id}")
def get_path_param_le(item_id: float = Path(..., valid=Num(le=3))):
    return item_id


@app.get("/path/param-lt-gt/{item_id}")
def get_path_param_lt_gt(item_id: float = Path(..., valid=Num(lt=3, gt=1))):
    return item_id


@app.get("/path/param-le-ge/{item_id}")
def get_path_param_le_ge(item_id: float = Path(..., valid=Num(le=3, ge=1))):
    return item_id


@app.get("/path/param-lt-int/{item_id}")
def get_path_param_lt_int(item_id: int = Path(..., valid=Num(lt=3))):
    return item_id


@app.get("/path/param-gt-int/{item_id}")
def get_path_param_gt_int(item_id: int = Path(..., valid=Num(gt=3))):
    return item_id


@app.get("/path/param-le-int/{item_id}")
def get_path_param_le_int(item_id: int = Path(..., valid=Num(le=3))):
    return item_id


@app.get("/path/param-ge-int/{item_id}")
def get_path_param_ge_int(item_id: int = Path(..., valid=Num(ge=3))):
    return item_id


@app.get("/path/param-lt-gt-int/{item_id}")
def get_path_param_lt_gt_int(item_id: int = Path(..., valid=Num(lt=3, gt=1))):
    return item_id


@app.get("/path/param-le-ge-int/{item_id}")
def get_path_param_le_ge_int(item_id: int = Path(..., valid=Num(le=3, ge=1))):
    return item_id


@app.get("/query")
def get_query(query):
    return f"foo bar {query}"


@app.get("/query/optional")
def get_query_optional(query=None):
    if query is None:
        return "foo bar"
    return f"foo bar {query}"


@app.get("/query/int")
def get_query_type(query: int):
    return f"foo bar {query}"


@app.get("/query/int/optional")
def get_query_type_optional(query: Optional[int] = None):
    if query is None:
        return "foo bar"
    return f"foo bar {query}"


@app.get("/query/int/default")
def get_query_type_int_default(query: int = 10):
    return f"foo bar {query}"


@app.get("/query/param")
def get_query_param(query=Query(None)):
    if query is None:
        return "foo bar"
    return f"foo bar {query}"


@app.get("/query/param-required")
def get_query_param_required(query=Query(...)):
    return f"foo bar {query}"


@app.get("/query/param-required/int")
def get_query_param_required_type(query: int = Query(...)):
    return f"foo bar {query}"


@app.get("/enum-status-code", status_code=http.HTTPStatus.CREATED)
def get_enum_status_code():
    return "foo bar"


client = TestClient(app)


def test_text_get():
    response = client.get("/text")
    assert response.status_code == 200, response.text
    assert response.json() == "Hello World"


def test_nonexistent():
    response = client.get("/nonexistent")
    assert response.status_code == 404, response.text
    assert response.json() == {"detail": "Not Found"}


response_not_valid_bool = {
    "detail": [
        {
            "loc": ["path", "item_id"],
            "msg": "value could not be parsed to a boolean",
            "type": "type_error.bool",
        }
    ]
}

response_not_valid_int = {
    "details": [
        {"loc": ["path_params", "item_id"], "msg": "Cast of `int` failed", "val": "2.7"}
    ]
}

response_not_valid_float = {
    "details": [
        {
            "loc": ["path_params", "item_id"],
            "msg": "Cast of `float` failed",
            "val": "True",
        }
    ]
}

response_at_least_3 = {
    "details": [
        {"loc": ["path_params", "item_id"], "msg": "Validation error", "val": "fo"}
    ]
}


response_at_least_2 = {
    "details": [
        {"loc": ["path_params", "item_id"], "msg": "Validation error", "val": "f"}
    ]
}


response_maximum_3 = {
    "details": [
        {"loc": ["path_params", "item_id"], "msg": "Validation error", "val": "foobar"}
    ]
}


response_greater_than_3 = {
    "details": [
        {"loc": ["path_params", "item_id"], "msg": "Validation error", "val": 2}
    ]
}


response_greater_than_0 = {
    "details": [
        {"loc": ["path_params", "item_id"], "msg": "Validation error", "val": 0.0}
    ]
}


response_greater_than_1 = {
    "details": [
        {"loc": ["path_params", "item_id"], "msg": "Validation error", "val": 0}
    ]
}


response_greater_than_equal_3 = {
    "details": [
        {"loc": ["path_params", "item_id"], "msg": "Validation error", "val": 2}
    ]
}


response_less_than_3 = {
    "details": [
        {"loc": ["path_params", "item_id"], "msg": "Validation error", "val": 4}
    ]
}


response_less_than_0 = {
    "details": [
        {"loc": ["path_params", "item_id"], "msg": "Validation error", "val": 0}
    ]
}


response_less_than_equal_3 = {
    "details": [
        {"loc": ["path_params", "item_id"], "msg": "Validation error", "val": 4}
    ]
}


@pytest.mark.parametrize(
    "path,expected_status,expected_response",
    [
        ("/path/foobar", 200, "foobar"),
        ("/path/str/foobar", 200, "foobar"),
        ("/path/str/42", 200, "42"),
        ("/path/str/True", 200, "True"),
        ("/path/int/2.7", 400, response_not_valid_int),
        ("/path/int/42", 200, 42),
        ("/path/int/2.7", 400, response_not_valid_int),
        ("/path/float/True", 400, response_not_valid_float),
        ("/path/float/42", 200, 42),
        ("/path/float/42.5", 200, 42.5),
        ("/path/param/foo", 200, "foo"),
        ("/path/param-required/foo", 200, "foo"),
        ("/path/param-minlength/foo", 200, "foo"),
        ("/path/param-minlength/fo", 400, response_at_least_3),
        ("/path/param-maxlength/foo", 200, "foo"),
        ("/path/param-maxlength/foobar", 400, response_maximum_3),
        ("/path/param-min_maxlength/foo", 200, "foo"),
        ("/path/param-min_maxlength/foobar", 400, response_maximum_3),
        ("/path/param-min_maxlength/f", 400, response_at_least_2),
        ("/path/param-gt/4", 200, 4),
        ("/path/param-gt/2", 400, response_greater_than_3),
        ("/path/param-gt0/0.05", 200, 0.05),
        ("/path/param-gt0/0", 400, response_greater_than_0),
        ("/path/param-ge/4", 200, 4),
        ("/path/param-ge/3", 200, 3),
        ("/path/param-ge/2", 400, response_greater_than_equal_3),
        ("/path/param-lt/4", 400, response_less_than_3),
        ("/path/param-lt/2", 200, 2),
        ("/path/param-lt0/-1", 200, -1),
        ("/path/param-lt0/0", 400, response_less_than_0),
        ("/path/param-le/4", 400, response_less_than_equal_3),
        ("/path/param-le/3", 200, 3),
        ("/path/param-le/2", 200, 2),
        ("/path/param-lt-gt/2", 200, 2),
        ("/path/param-lt-gt/4", 400, response_less_than_3),
        ("/path/param-lt-gt/0", 400, response_greater_than_1),
        ("/path/param-le-ge/2", 200, 2),
        ("/path/param-le-ge/1", 200, 1),
        ("/path/param-le-ge/3", 200, 3),
        ("/path/param-le-ge/4", 400, response_less_than_equal_3),
        ("/path/param-lt-int/2", 200, 2),
        ("/path/param-lt-int/4", 400, response_less_than_3),
        ("/path/param-lt-int/2.7", 400, response_not_valid_int),
        ("/path/param-gt-int/4", 200, 4),
        ("/path/param-gt-int/2", 400, response_greater_than_3),
        ("/path/param-gt-int/2.7", 400, response_not_valid_int),
        ("/path/param-le-int/4", 400, response_less_than_equal_3),
        ("/path/param-le-int/3", 200, 3),
        ("/path/param-le-int/2", 200, 2),
        ("/path/param-le-int/2.7", 400, response_not_valid_int),
        ("/path/param-ge-int/42", 200, 42),
        ("/path/param-ge-int/3", 200, 3),
        ("/path/param-ge-int/2", 400, response_greater_than_equal_3),
        ("/path/param-ge-int/2.7", 400, response_not_valid_int),
        ("/path/param-lt-gt-int/2", 200, 2),
        ("/path/param-lt-gt-int/4", 400, response_less_than_3),
        ("/path/param-lt-gt-int/0", 400, response_greater_than_1),
        ("/path/param-lt-gt-int/2.7", 400, response_not_valid_int),
        ("/path/param-le-ge-int/2", 200, 2),
        ("/path/param-le-ge-int/1", 200, 1),
        ("/path/param-le-ge-int/3", 200, 3),
        ("/path/param-le-ge-int/4", 400, response_less_than_equal_3),
        ("/path/param-le-ge-int/2.7", 400, response_not_valid_int),
    ],
)
def test_get_path(path, expected_status, expected_response):
    response = client.get(path)
    assert response.status_code == expected_status
    assert response.json() == expected_response



from squall import HTTPException, Squall
from squall.exceptions import RequestPayloadValidationError
from squall.responses import PrettyJSONResponse
from squall.testclient import TestClient


def http_exception_handler(request, exception):
    return PrettyJSONResponse({"exception": "http-exception"})


def request_validation_exception_handler(request, exception):
    return PrettyJSONResponse({"exception": "request-validation"})


app = Squall(
    exception_handlers={
        HTTPException: http_exception_handler,
        RequestPayloadValidationError: request_validation_exception_handler,
    }
)

client = TestClient(app)


@app.get("/http-exception")
def route_with_http_exception():
    raise HTTPException(status_code=400)


@app.get("/request-validation/{param}/")
def route_with_request_validation_exception(param: int):
    pass  # pragma: no cover


def test_override_http_exception():
    response = client.get("/http-exception")
    assert response.status_code == 200
    assert response.json() == {"exception": "http-exception"}


def test_override_request_validation_exception():
    response = client.get("/request-validation/invalid")
    assert response.status_code == 400
    assert response.json() == {
        "details": [
            {
                "loc": ["path_params", "param"],
                "msg": "Cast of `int` failed",
                "val": "invalid",
            }
        ]
    }



import typing
from decimal import Decimal

import pytest
from squall.params import Cookie, Header, Num, Path, Query, Str


def test_no_args(app, client):
    @app.get("/foo")
    def foo():
        return {}

    response = client.get("/foo")
    assert response.status_code == 200, response.text
    assert response.json() == {}


def test_no_defaults_no_annotation(app, client):
    @app.get("/no_defaults/no_annotations/{from_path}")
    def foo(
        from_path, from_qs=Query(...), from_header=Header(...), from_cookie=Cookie(...)
    ):
        return {
            "from_path": from_path,
            "from_path_type": type(from_path).__name__,
            "from_qs": from_qs,
            "from_qs_type": type(from_qs).__name__,
            "from_header": from_header,
            "from_header_type": type(from_header).__name__,
            "from_cookie": from_cookie,
            "from_cookie_type": type(from_cookie).__name__,
        }

    response = client.get("/no_defaults/no_annotations/some_param")
    assert response.status_code == 400, response.text
    assert response.json() == {
        "details": [
            {"loc": ["query_params", "from_qs"], "msg": "Mandatory field missed"},
            {"loc": ["headers", "from_header"], "msg": "Mandatory field missed"},
            {"loc": ["cookies", "from_cookie"], "msg": "Mandatory field missed"},
        ]
    }

    response = client.get(
        "/no_defaults/no_annotations/some_param?from_qs=qs_value",
        headers={"from_header": "header_value"},
        cookies={"from_cookie": "cookie_value"},
    )
    assert response.status_code == 200, response.text
    assert response.json() == {
        "from_path": "some_param",
        "from_path_type": "str",
        "from_qs": "qs_value",
        "from_qs_type": "str",
        "from_header": "header_value",
        "from_header_type": "str",
        "from_cookie": "cookie_value",
        "from_cookie_type": "str",
    }


def test_defaults_no_annotation(app, client):
    @app.get("/defaults/no_annotations")
    def foo(from_qs=Query(None), from_header=Header(None), from_cookie=Cookie(None)):
        return {
            "from_qs": from_qs,
            "from_qs_type": type(from_qs).__name__,
            "from_header": from_header,
            "from_header_type": type(from_header).__name__,
            "from_cookie": from_cookie,
            "from_cookie_type": type(from_cookie).__name__,
        }

    response = client.get("/defaults/no_annotations")
    assert response.status_code == 200, response.text
    assert response.json() == {
        "from_qs": None,
        "from_qs_type": "NoneType",
        "from_header": None,
        "from_header_type": "NoneType",
        "from_cookie": None,
        "from_cookie_type": "NoneType",
    }


def test_parameter_from_another_alias_name(app, client):
    @app.get("/from_another/alias_name/{alias_param}")
    def foo(
        param=Path(..., alias="alias_param"),
        qs=Query(..., alias="alias_qs"),
        header=Header(..., alias="alias_header"),
        cookie=Cookie(..., alias="alias_cookie"),
    ):
        return {
            "param": param,
            "param_type": type(param).__name__,
            "qs": qs,
            "qs_type": type(qs).__name__,
            "header": header,
            "header_type": type(header).__name__,
            "cookie": cookie,
            "cookie_type": type(cookie).__name__,
        }

    response = client.get(
        "/from_another/alias_name/some_param?alias_qs=qs_value",
        headers={"alias_header": "header_value"},
        cookies={"alias_cookie": "cookie_value"},
    )
    assert response.status_code == 200, response.text
    assert response.json() == {
        "param": "some_param",
        "param_type": "str",
        "qs": "qs_value",
        "qs_type": "str",
        "header": "header_value",
        "header_type": "str",
        "cookie": "cookie_value",
        "cookie_type": "str",
    }


def test_no_defaults_cast_from_annotation_positive(app, client):
    @app.get("/no_defaults/cast_from_annotation/{from_path}")
    def foo(
        from_path: int,
        from_qs: float = Query(...),
        from_header: str = Header(...),
        from_cookie: bytes = Cookie(...),
    ):
        return {
            "from_path": from_path,
            "from_path_type": type(from_path).__name__,
            "from_qs": from_qs,
            "from_qs_type": type(from_qs).__name__,
            "from_header": from_header,
            "from_header_type": type(from_header).__name__,
            "from_cookie": from_cookie,
            "from_cookie_type": type(from_cookie).__name__,
        }

    response = client.get(
        "/no_defaults/cast_from_annotation/123?from_qs=321",
        headers={"from_header": "header_value"},
        cookies={"from_cookie": "cookie_value"},
    )
    assert response.status_code == 200, response.text
    assert response.json() == {
        "from_path": 123,
        "from_path_type": "int",
        "from_qs": 321.0,
        "from_qs_type": "float",
        "from_header": "header_value",
        "from_header_type": "str",
        "from_cookie": "cookie_value",
        "from_cookie_type": "bytes",
    }


def test_validate_any_string_positive(app, client):
    @app.get("/validate/any_string/{from_path}")
    def foo(
        from_path: str = Path(..., valid=Str(min_len=5, max_len=20)),
        from_qs: str = Query(..., valid=Str(min_len=5, max_len=20)),
        from_header: bytes = Header(..., valid=Str(min_len=5, max_len=20)),
        from_cookie: bytes = Cookie(..., valid=Str(min_len=5, max_len=20)),
    ):
        return {
            "from_path": from_path,
            "from_path_type": type(from_path).__name__,
            "from_qs": from_qs,
            "from_qs_type": type(from_qs).__name__,
            "from_header": from_header,
            "from_header_type": type(from_header).__name__,
            "from_cookie": from_cookie,
            "from_cookie_type": type(from_cookie).__name__,
        }

    response = client.get(
        "/validate/any_string/12345?from_qs=32132",
        headers={"from_header": "header_value"},
        cookies={"from_cookie": "cookie_value"},
    )
    assert response.status_code == 200, response.text
    assert response.json() == {
        "from_path": "12345",
        "from_path_type": "str",
        "from_qs": "32132",
        "from_qs_type": "str",
        "from_header": "header_value",
        "from_header_type": "bytes",
        "from_cookie": "cookie_value",
        "from_cookie_type": "bytes",
    }


@pytest.mark.parametrize(
    "min_len,max_len",
    [
        (6, None),
        (None, 4),
        (1, 4),
    ],
)
def test_validate_any_string_negative(app, client, min_len, max_len):
    @app.get("/validate/any_string/{from_path}")
    def foo(
        from_path: str = Path(..., valid=Str(min_len=min_len, max_len=max_len)),
        from_qs: str = Query(..., valid=Str(min_len=min_len, max_len=max_len)),
        from_header: bytes = Header(..., valid=Str(min_len=min_len, max_len=max_len)),
        from_cookie: bytes = Cookie(..., valid=Str(min_len=min_len, max_len=max_len)),
    ):
        return {
            "from_path": from_path,
            "from_path_type": type(from_path).__name__,
            "from_qs": from_qs,
            "from_qs_type": type(from_qs).__name__,
            "from_header": from_header,
            "from_header_type": type(from_header).__name__,
            "from_cookie": from_cookie,
            "from_cookie_type": type(from_cookie).__name__,
        }

    response = client.get(
        "/validate/any_string/12345?from_qs=12345",
        headers={"from_header": "12345"},
        cookies={"from_cookie": "12345"},
    )
    assert response.status_code == 400, response.text
    assert response.json() == {
        "details": [
            {
                "loc": ["path_params", "from_path"],
                "msg": "Validation error",
                "val": "12345",
            },
            {
                "loc": ["query_params", "from_qs"],
                "msg": "Validation error",
                "val": "12345",
            },
            {
                "loc": ["headers", "from_header"],
                "msg": "Validation error",
                "val": "12345",
            },
            {
                "loc": ["cookies", "from_cookie"],
                "msg": "Validation error",
                "val": "12345",
            },
        ]
    }


def test_validate_numeric_positive(app, client):
    @app.get("/validate/numeric/{from_path}")
    def foo(
        from_path: int = Path(..., valid=Num(ge=5, le=20)),
        from_qs: float = Query(..., valid=Num(ge=5, le=20)),
        from_header: Decimal = Header(..., valid=Num(ge=5, le=20)),
        from_cookie: int = Cookie(..., valid=Num(ge=5, le=20)),
    ):
        return {
            "from_path": from_path,
            "from_path_type": type(from_path).__name__,
            "from_qs": from_qs,
            "from_qs_type": type(from_qs).__name__,
            "from_header": from_header,
            "from_header_type": type(from_header).__name__,
            "from_cookie": from_cookie,
            "from_cookie_type": type(from_cookie).__name__,
        }

    response = client.get(
        "/validate/numeric/15?from_qs=16",
        headers={"from_header": "14"},
        cookies={"from_cookie": "17"},
    )
    assert response.status_code == 200, response.text
    assert response.json() == {
        "from_path": 15,
        "from_path_type": "int",
        "from_qs": 16.0,
        "from_qs_type": "float",
        "from_header": Decimal(14),
        "from_header_type": "Decimal",
        "from_cookie": 17,
        "from_cookie_type": "int",
    }


@pytest.mark.parametrize(
    "ge,le",
    [
        (20, None),
        (None, 5),
        (1, 4),
    ],
)
def test_validate_numeric_negative(app, client, ge, le):
    @app.get("/validate/numeric/{from_path}")
    def foo(
        from_path: int = Path(..., valid=Num(ge=ge, le=le)),
        from_qs: float = Query(..., valid=Num(ge=ge, le=le)),
        from_header: Decimal = Header(..., valid=Num(ge=ge, le=le)),
        from_cookie: int = Cookie(..., valid=Num(ge=ge, le=le)),
    ):
        return {
            "from_path": from_path,
            "from_path_type": type(from_path).__name__,
            "from_qs": from_qs,
            "from_qs_type": type(from_qs).__name__,
            "from_header": from_header,
            "from_header_type": type(from_header).__name__,
            "from_cookie": from_cookie,
            "from_cookie_type": type(from_cookie).__name__,
        }

    response = client.get(
        "/validate/numeric/15?from_qs=16",
        headers={"from_header": "14"},
        cookies={"from_cookie": "17"},
    )
    assert response.status_code == 400, response.text
    assert response.json() == {
        "details": [
            {"loc": ["path_params", "from_path"], "msg": "Validation error", "val": 15},
            {
                "loc": ["query_params", "from_qs"],
                "msg": "Validation error",
                "val": 16.0,
            },
            {"loc": ["headers", "from_header"], "msg": "Validation error", "val": 14.0},
            {"loc": ["cookies", "from_cookie"], "msg": "Validation error", "val": 17},
        ]
    }


def test_list_annotation(app, client):
    @app.get("/validate/any_string/{from_path}")
    def foo(
        from_qs: typing.List[str] = Query(...),
        from_qs2: typing.Optional[typing.List[bytes]] = Query(...),
    ):
        return {
            "from_qs": from_qs,
            "from_qs_type": type(from_qs).__name__,
            "from_qs_element_type": type(from_qs[0]).__name__,
            "from_qs2": from_qs2,
            "from_qs2_type": type(from_qs2).__name__,
            "from_qs2_element_type": type(from_qs2[0]).__name__,
        }

    response = client.get(
        "/validate/any_string/12345?"
        "from_qs=11111&from_qs=22222&"
        "from_qs2=11111&from_qs2=22222"
    )
    assert response.status_code == 200, response.text
    assert response.json() == {
        "from_qs": ["11111", "22222"],
        "from_qs2": ["11111", "22222"],
        "from_qs2_element_type": "bytes",
        "from_qs2_type": "list",
        "from_qs_element_type": "str",
        "from_qs_type": "list",
    }



import asyncio
import traceback
from typing import List, TypeVar

from squall.types import AnyFunc, Receive, Scope, Send

_T = TypeVar("_T")


class LifespanContext:
    def __init__(self, on_startup: List[AnyFunc], on_shutdown: List[AnyFunc]):
        self._on_startup = on_startup
        self._on_shutdown = on_shutdown

    @staticmethod
    async def _run_handlers(handlers: List[AnyFunc]) -> None:
        for handler in handlers:
            if asyncio.iscoroutinefunction(handler):
                await handler()
            else:
                handler()

    async def __aenter__(self) -> None:
        await self._run_handlers(self._on_startup)

    async def __aexit__(self, *exc_info: object) -> None:
        await self._run_handlers(self._on_shutdown)

    def __call__(self: _T, app: object) -> _T:
        return self


async def lifespan(
    ctx: LifespanContext,
    scope: Scope,
    receive: Receive,
    send: Send,
) -> None:
    """
    Handle ASGI lifespan messages, which allows us to manage application
    startup and shutdown events.
    """
    started = False
    await receive()
    try:
        async with ctx:
            await send({"type": "lifespan.startup.complete"})
            started = True
            await receive()
    except BaseException:
        exc_text = traceback.format_exc()
        if started:
            await send({"type": "lifespan.shutdown.failed", "message": exc_text})
        else:
            await send({"type": "lifespan.startup.failed", "message": exc_text})
        raise
    else:
        await send({"type": "lifespan.shutdown.complete"})



from dataclasses import dataclass
from typing import Any, Optional

from squall.params import Body


@dataclass
class RequestField:
    name: str
    model: Any
    settings: Optional[Body] = None


@dataclass
class ResponseField:
    model: Any



import os
import shutil
from http.server import HTTPServer, SimpleHTTPRequestHandler
from multiprocessing import Pool
from pathlib import Path
from typing import Dict, List, Optional, Tuple

import mkdocs.commands.build
import mkdocs.commands.serve
import mkdocs.config
import mkdocs.utils
import typer
import yaml

app = typer.Typer()

mkdocs_name = "mkdocs.yml"

missing_translation_snippet = """
{!../../../docs/missing-translation.md!}
"""

docs_path = Path("docs")
en_docs_path = Path("docs/en")
en_config_path: Path = en_docs_path / mkdocs_name


def get_en_config() -> dict:
    return mkdocs.utils.yaml_load(en_config_path.read_text(encoding="utf-8"))


def get_lang_paths():
    return sorted(docs_path.iterdir())


def lang_callback(lang: Optional[str]):
    if lang is None:
        return
    if not lang.isalpha() or len(lang) != 2:
        typer.echo("Use a 2 letter language code, like: es")
        raise typer.Abort()
    lang = lang.lower()
    return lang


def complete_existing_lang(incomplete: str):
    lang_path: Path
    for lang_path in get_lang_paths():
        if lang_path.is_dir() and lang_path.name.startswith(incomplete):
            yield lang_path.name


def get_base_lang_config(lang: str):
    en_config = get_en_config()
    squall_url_base = "https://squall.mtag.dev/"
    new_config = en_config.copy()
    new_config["site_url"] = en_config["site_url"] + f"{lang}/"
    new_config["theme"]["logo"] = squall_url_base + en_config["theme"]["logo"]
    new_config["theme"]["favicon"] = squall_url_base + en_config["theme"]["favicon"]
    new_config["theme"]["language"] = lang
    new_config["nav"] = en_config["nav"][:2]
    extra_css = []
    css: str
    for css in en_config["extra_css"]:
        if css.startswith("http"):
            extra_css.append(css)
        else:
            extra_css.append(squall_url_base + css)
    new_config["extra_css"] = extra_css

    extra_js = []
    js: str
    for js in en_config["extra_javascript"]:
        if js.startswith("http"):
            extra_js.append(js)
        else:
            extra_js.append(squall_url_base + js)
    new_config["extra_javascript"] = extra_js
    return new_config


@app.command()
def new_lang(lang: str = typer.Argument(..., callback=lang_callback)):
    """
    Generate a new docs translation directory for the language LANG.

    LANG should be a 2-letter language code, like: en, es, de, pt, etc.
    """
    new_path: Path = Path("docs") / lang
    if new_path.exists():
        typer.echo(f"The language was already created: {lang}")
        raise typer.Abort()
    new_path.mkdir()
    new_config = get_base_lang_config(lang)
    new_config_path: Path = Path(new_path) / mkdocs_name
    new_config_path.write_text(
        yaml.dump(new_config, sort_keys=False, width=200, allow_unicode=True),
        encoding="utf-8",
    )
    new_config_docs_path: Path = new_path / "docs"
    new_config_docs_path.mkdir()
    en_index_path: Path = en_docs_path / "docs" / "index.md"
    new_index_path: Path = new_config_docs_path / "index.md"
    en_index_content = en_index_path.read_text(encoding="utf-8")
    new_index_content = f"{missing_translation_snippet}\n\n{en_index_content}"
    new_index_path.write_text(new_index_content, encoding="utf-8")
    new_overrides_gitignore_path = new_path / "overrides" / ".gitignore"
    new_overrides_gitignore_path.parent.mkdir(parents=True, exist_ok=True)
    new_overrides_gitignore_path.write_text("")
    typer.secho(f"Successfully initialized: {new_path}", color=typer.colors.GREEN)
    update_languages(lang=None)


@app.command()
def build_lang(
    lang: str = typer.Argument(
        ..., callback=lang_callback, autocompletion=complete_existing_lang
    )
):
    """
    Build the docs for a language, filling missing pages with translation notifications.
    """
    lang_path: Path = Path("docs") / lang
    if not lang_path.is_dir():
        typer.echo(f"The language translation doesn't seem to exist yet: {lang}")
        raise typer.Abort()
    typer.echo(f"Building docs for: {lang}")
    build_dir_path = Path("docs_build")
    build_dir_path.mkdir(exist_ok=True)
    build_lang_path = build_dir_path / lang
    en_lang_path = Path("docs/en")
    site_path = Path("site").absolute()
    if lang == "en":
        dist_path = site_path
    else:
        dist_path: Path = site_path / lang
    shutil.rmtree(build_lang_path, ignore_errors=True)
    shutil.copytree(lang_path, build_lang_path)
    shutil.copytree(en_docs_path / "data", build_lang_path / "data")
    overrides_src = en_docs_path / "overrides"
    overrides_dest = build_lang_path / "overrides"
    for path in overrides_src.iterdir():
        dest_path = overrides_dest / path.name
        if not dest_path.exists():
            shutil.copy(path, dest_path)
    en_config_path: Path = en_lang_path / mkdocs_name
    en_config: dict = mkdocs.utils.yaml_load(en_config_path.read_text(encoding="utf-8"))
    nav = en_config["nav"]
    lang_config_path: Path = lang_path / mkdocs_name
    lang_config: dict = mkdocs.utils.yaml_load(
        lang_config_path.read_text(encoding="utf-8")
    )
    lang_nav = lang_config["nav"]
    # Exclude first 2 entries Squall and Languages, for custom handling
    use_nav = nav[2:]
    lang_use_nav = lang_nav[2:]
    file_to_nav = get_file_to_nav_map(use_nav)
    sections = get_sections(use_nav)
    lang_file_to_nav = get_file_to_nav_map(lang_use_nav)
    use_lang_file_to_nav = get_file_to_nav_map(lang_use_nav)
    for file in file_to_nav:
        file_path = Path(file)
        lang_file_path: Path = build_lang_path / "docs" / file_path
        en_file_path: Path = en_lang_path / "docs" / file_path
        lang_file_path.parent.mkdir(parents=True, exist_ok=True)
        if not lang_file_path.is_file():
            en_text = en_file_path.read_text(encoding="utf-8")
            lang_text = get_text_with_translate_missing(en_text)
            lang_file_path.write_text(lang_text, encoding="utf-8")
            file_key = file_to_nav[file]
            use_lang_file_to_nav[file] = file_key
            if file_key:
                composite_key = ()
                new_key = ()
                for key_part in file_key:
                    composite_key += (key_part,)
                    key_first_file = sections[composite_key]
                    if key_first_file in lang_file_to_nav:
                        new_key = lang_file_to_nav[key_first_file]
                    else:
                        new_key += (key_part,)
                use_lang_file_to_nav[file] = new_key
    key_to_section = {(): []}
    for file, orig_file_key in file_to_nav.items():
        if file in use_lang_file_to_nav:
            file_key = use_lang_file_to_nav[file]
        else:
            file_key = orig_file_key
        section = get_key_section(key_to_section=key_to_section, key=file_key)
        section.append(file)
    new_nav = key_to_section[()]
    export_lang_nav = [lang_nav[0], nav[1]] + new_nav
    lang_config["nav"] = export_lang_nav
    build_lang_config_path: Path = build_lang_path / mkdocs_name
    build_lang_config_path.write_text(
        yaml.dump(lang_config, sort_keys=False, width=200, allow_unicode=True),
        encoding="utf-8",
    )
    current_dir = os.getcwd()
    os.chdir(build_lang_path)
    mkdocs.commands.build.build(mkdocs.config.load_config(site_dir=str(dist_path)))
    os.chdir(current_dir)
    typer.secho(f"Successfully built docs for: {lang}", color=typer.colors.GREEN)


def generate_readme_content():
    en_index = en_docs_path / "docs" / "index.md"
    return en_index.read_text("utf-8")


@app.command()
def generate_readme():
    """
    Generate README.md content from main index.md
    """
    typer.echo("Generating README")
    readme_path = Path("README.md")
    new_content = generate_readme_content()
    readme_path.write_text(new_content, encoding="utf-8")


@app.command()
def verify_readme():
    """
    Verify README.md content from main index.md
    """
    typer.echo("Verifying README")
    readme_path = Path("README.md")
    generated_content = generate_readme_content()
    readme_content = readme_path.read_text("utf-8")
    if generated_content != readme_content:
        typer.secho(
            "README.md outdated from the latest index.md", color=typer.colors.RED
        )
        raise typer.Abort()
    typer.echo("Valid README ✅")


@app.command()
def build_all():
    """
    Build mkdocs site for en, and then build each language inside, end result is located
    at directory ./site/ with each language inside.
    """
    site_path = Path("site").absolute()
    update_languages(lang=None)
    current_dir = os.getcwd()
    os.chdir(en_docs_path)
    typer.echo("Building docs for: en")
    mkdocs.commands.build.build(mkdocs.config.load_config(site_dir=str(site_path)))
    os.chdir(current_dir)
    langs = [
        lang.name for lang in get_lang_paths() if lang != en_docs_path and lang.is_dir()
    ]

    cpu_count = os.cpu_count() or 1
    with Pool(cpu_count * 2) as p:
        p.map(build_lang, langs)


def update_single_lang(lang: str):
    lang_path = docs_path / lang
    typer.echo(f"Updating {lang_path.name}")
    update_config(lang_path.name)


@app.command()
def update_languages(
    lang: str = typer.Argument(
        None, callback=lang_callback, autocompletion=complete_existing_lang
    )
):
    """
    Update the mkdocs.yml file Languages section including all the available languages.

    The LANG argument is a 2-letter language code. If it's not provided, update all the
    mkdocs.yml files (for all the languages).
    """
    if lang is None:
        for lang_path in get_lang_paths():
            if lang_path.is_dir():
                update_single_lang(lang_path.name)
    else:
        update_single_lang(lang)


@app.command()
def serve():
    """
    A quick server to preview a built site with translations.

    For development, prefer the command live (or just mkdocs serve).

    This is here only to preview a site with translations already built.

    Make sure you run the build-all command first.
    """
    typer.echo("Warning: this is a very simple server.")
    typer.echo("For development, use the command live instead.")
    typer.echo("This is here only to preview a site with translations already built.")
    typer.echo("Make sure you run the build-all command first.")
    os.chdir("site")
    server_address = ("", 8008)
    server = HTTPServer(server_address, SimpleHTTPRequestHandler)
    typer.echo("Serving at: http://127.0.0.1:8008")
    server.serve_forever()


@app.command()
def live(
    lang: str = typer.Argument(
        None, callback=lang_callback, autocompletion=complete_existing_lang
    )
):
    """
    Serve with livereload a docs site for a specific language.

    This only shows the actual translated files, not the placeholders created with
    build-all.

    Takes an optional LANG argument with the name of the language to serve, by default
    en.
    """
    if lang is None:
        lang = "en"
    lang_path: Path = docs_path / lang
    os.chdir(lang_path)
    mkdocs.commands.serve.serve(dev_addr="127.0.0.1:8008")


def update_config(lang: str):
    lang_path: Path = docs_path / lang
    config_path = lang_path / mkdocs_name
    current_config: dict = mkdocs.utils.yaml_load(
        config_path.read_text(encoding="utf-8")
    )
    if lang == "en":
        config = get_en_config()
    else:
        config = get_base_lang_config(lang)
        config["nav"] = current_config["nav"]
        config["theme"]["language"] = current_config["theme"]["language"]
    languages = [{"en": "/"}]
    alternate: List[Dict[str, str]] = config["extra"].get("alternate", [])
    alternate_dict = {alt["link"]: alt["name"] for alt in alternate}
    new_alternate: List[Dict[str, str]] = []
    for lang_path in get_lang_paths():
        if lang_path.name == "en" or not lang_path.is_dir():
            continue
        name = lang_path.name
        languages.append({name: f"/{name}/"})
    for lang_dict in languages:
        name = list(lang_dict.keys())[0]
        url = lang_dict[name]
        if url not in alternate_dict:
            new_alternate.append({"link": url, "name": name})
        else:
            use_name = alternate_dict[url]
            new_alternate.append({"link": url, "name": use_name})
    config["nav"][1] = {"Languages": languages}
    config["extra"]["alternate"] = new_alternate
    config_path.write_text(
        yaml.dump(config, sort_keys=False, width=200, allow_unicode=True),
        encoding="utf-8",
    )


def get_key_section(
    *, key_to_section: Dict[Tuple[str, ...], list], key: Tuple[str, ...]
) -> list:
    if key in key_to_section:
        return key_to_section[key]
    super_key = key[:-1]
    title = key[-1]
    super_section = get_key_section(key_to_section=key_to_section, key=super_key)
    new_section = []
    super_section.append({title: new_section})
    key_to_section[key] = new_section
    return new_section


def get_text_with_translate_missing(text: str) -> str:
    lines = text.splitlines()
    lines.insert(1, missing_translation_snippet)
    return "\n".join(lines)


def get_file_to_nav_map(nav: list) -> Dict[str, Tuple[str, ...]]:
    file_to_nav = {}
    for item in nav:
        if type(item) is str:
            file_to_nav[item] = tuple()
        elif type(item) is dict:
            item_key = list(item.keys())[0]
            sub_nav = item[item_key]
            sub_file_to_nav = get_file_to_nav_map(sub_nav)
            for k, v in sub_file_to_nav.items():
                file_to_nav[k] = (item_key,) + v
    return file_to_nav


def get_sections(nav: list) -> Dict[Tuple[str, ...], str]:
    sections = {}
    for item in nav:
        if type(item) is str:
            continue
        elif type(item) is dict:
            item_key = list(item.keys())[0]
            sub_nav = item[item_key]
            sections[(item_key,)] = sub_nav[0]
            sub_sections = get_sections(sub_nav)
            for k, v in sub_sections.items():
                new_key = (item_key,) + k
                sections[new_key] = v
    return sections


if __name__ == "__main__":
    app()



from typing import Optional

from squall.params import Cookie, Header, Num, Path, Query, Str
from squall.routing_.utils import get_handler_head_params


def test_get_handler_args_no_args():
    def handler():
        pass

    assert get_handler_head_params(handler) == []


def test_get_handler_args_no_annotation():
    def handler(a, b=Query(...), c=Path(...), d=Header(...), e=Cookie(...)):
        pass

    params = get_handler_head_params(handler)
    assert params[0].convertor == "str"
    assert params[0].default == Ellipsis
    assert params[0].statements == {}
    assert params[0].name == "a"
    assert params[0].alias == "a"
    assert params[0].source == "path_params"
    assert params[0].validate is None

    assert params[1].convertor == "str"
    assert params[1].default == Ellipsis
    assert params[1].statements == {}
    assert params[1].name == "b"
    assert params[1].alias == "b"
    assert params[1].source == "query_params"
    assert params[1].validate is None

    assert params[2].convertor == "str"
    assert params[2].default == Ellipsis
    assert params[2].statements == {}
    assert params[2].name == "c"
    assert params[2].alias == "c"
    assert params[2].source == "path_params"
    assert params[2].validate is None

    assert params[3].convertor == "str"
    assert params[3].default == Ellipsis
    assert params[3].statements == {}
    assert params[3].name == "d"
    assert params[3].alias == "d"
    assert params[3].source == "headers"
    assert params[3].validate is None

    assert params[4].convertor == "str"
    assert params[4].default == Ellipsis
    assert params[4].statements == {}
    assert params[4].name == "e"
    assert params[4].alias == "e"
    assert params[4].source == "cookies"
    assert params[4].validate is None


def test_get_handler_args_alias():
    def handler(
        b=Query(..., alias="from_b"),
        c=Path(..., alias="from_c"),
        d=Header(..., alias="from_d"),
        e=Cookie(..., alias="from_e"),
    ):
        pass

    params = get_handler_head_params(handler)
    assert params[0].convertor == "str"
    assert params[0].default == Ellipsis
    assert params[0].statements == {}
    assert params[0].name == "b"
    assert params[0].alias == "from_b"
    assert params[0].source == "query_params"
    assert params[0].validate is None

    assert params[1].convertor == "str"
    assert params[1].default == Ellipsis
    assert params[1].statements == {}
    assert params[1].name == "c"
    assert params[1].alias == "from_c"
    assert params[1].source == "path_params"
    assert params[1].validate is None

    assert params[2].convertor == "str"
    assert params[2].default == Ellipsis
    assert params[2].statements == {}
    assert params[2].name == "d"
    assert params[2].alias == "from_d"
    assert params[2].source == "headers"
    assert params[2].validate is None

    assert params[3].convertor == "str"
    assert params[3].default == Ellipsis
    assert params[3].statements == {}
    assert params[3].name == "e"
    assert params[3].alias == "from_e"
    assert params[3].source == "cookies"
    assert params[3].validate is None


def test_get_handler_args_annotations():
    def handler(b: int, c: str, d: bytes, e: float):
        pass

    params = get_handler_head_params(handler)
    assert params[0].convertor == "int"
    assert params[0].default == Ellipsis
    assert params[0].statements == {}
    assert params[0].name == "b"
    assert params[0].source == "path_params"
    assert params[0].validate is None

    assert params[1].convertor == "str"
    assert params[1].default == Ellipsis
    assert params[1].statements == {}
    assert params[1].name == "c"
    assert params[1].source == "path_params"
    assert params[1].validate is None

    assert params[2].convertor == "bytes"
    assert params[2].default == Ellipsis
    assert params[2].statements == {}
    assert params[2].name == "d"
    assert params[2].source == "path_params"
    assert params[2].validate is None

    assert params[3].convertor == "float"
    assert params[3].default == Ellipsis
    assert params[3].statements == {}
    assert params[3].name == "e"
    assert params[3].source == "path_params"
    assert params[3].validate is None


def test_get_handler_args_direct_defaults():
    def handler(
        a: Optional[str] = None,
        b: int = 1,
        c: str = "Hey",
        d: bytes = b"Hey",
        e: float = 3.14,
    ):
        pass

    params = get_handler_head_params(handler)
    assert params[0].convertor == "str"
    assert params[0].default is None
    assert params[0].statements == {}
    assert params[0].name == "a"
    assert params[0].source == "path_params"
    assert params[0].validate is None

    assert params[1].convertor == "int"
    assert params[1].default == 1
    assert params[1].statements == {}
    assert params[1].name == "b"
    assert params[1].source == "path_params"
    assert params[1].validate is None

    assert params[2].convertor == "str"
    assert params[2].default == "Hey"
    assert params[2].statements == {}
    assert params[2].name == "c"
    assert params[2].source == "path_params"
    assert params[2].validate is None

    assert params[3].convertor == "bytes"
    assert params[3].default == b"Hey"
    assert params[3].statements == {}
    assert params[3].name == "d"
    assert params[3].source == "path_params"
    assert params[3].validate is None

    assert params[4].convertor == "float"
    assert params[4].default == 3.14
    assert params[4].statements == {}
    assert params[4].name == "e"
    assert params[4].source == "path_params"
    assert params[4].validate is None


def test_get_handler_args_assigned_instance_defaults():
    def handler(
        a: Optional[str] = Path(None),
        b: int = Path(1),
        c: str = Query("Hey"),
        d: bytes = Header(b"Hey"),
        e: float = Cookie(3.14),
    ):
        pass

    params = get_handler_head_params(handler)
    assert params[0].convertor == "str"
    assert params[0].default is None
    assert params[0].statements == {}
    assert params[0].name == "a"
    assert params[0].source == "path_params"
    assert params[0].validate is None

    assert params[1].convertor == "int"
    assert params[1].default == 1
    assert params[1].statements == {}
    assert params[1].name == "b"
    assert params[1].source == "path_params"
    assert params[1].validate is None

    assert params[2].convertor == "str"
    assert params[2].default == "Hey"
    assert params[2].statements == {}
    assert params[2].name == "c"
    assert params[2].source == "query_params"
    assert params[2].validate is None

    assert params[3].convertor == "bytes"
    assert params[3].default == b"Hey"
    assert params[3].statements == {}
    assert params[3].name == "d"
    assert params[3].source == "headers"
    assert params[3].validate is None

    assert params[4].convertor == "float"
    assert params[4].default == 3.14
    assert params[4].statements == {}
    assert params[4].name == "e"
    assert params[4].source == "cookies"
    assert params[4].validate is None


def test_get_handler_args_validation_parameters():
    def handler(
        a: Optional[str] = Path(None, valid=Str(min_len=2)),
        b: int = Path(1, valid=Num(ge=1, le=2)),
        c: str = Query("Hey", valid=Str(min_len=2, max_len=5)),
        d: bytes = Header(b"Hey", valid=Str(min_len=2, max_len=5)),
        e: float = Cookie(3.14, valid=Num(ge=3.14, le=3.15)),
    ):
        pass

    params = get_handler_head_params(handler)
    assert params[0].convertor == "str"
    assert params[0].default is None
    assert params[0].statements["min_len"] == 2
    assert params[0].name == "a"
    assert params[0].source == "path_params"
    assert params[0].validate == "string"

    assert params[1].convertor == "int"
    assert params[1].default == 1
    assert params[1].statements == {"ge": 1, "gt": None, "le": 2, "lt": None}
    assert params[1].name == "b"
    assert params[1].source == "path_params"
    assert params[1].validate == "numeric"

    assert params[2].convertor == "str"
    assert params[2].default == "Hey"
    assert params[2].statements == {"max_len": 5, "min_len": 2}
    assert params[2].name == "c"
    assert params[2].source == "query_params"
    assert params[2].validate == "string"

    assert params[3].convertor == "bytes"
    assert params[3].default == b"Hey"
    assert params[3].statements == {"max_len": 5, "min_len": 2}
    assert params[3].name == "d"
    assert params[3].source == "headers"
    assert params[3].validate == "string"

    assert params[4].convertor == "float"
    assert params[4].default == 3.14
    assert params[4].statements == {"ge": 3.14, "gt": None, "le": 3.15, "lt": None}
    assert params[4].name == "e"
    assert params[4].source == "cookies"
    assert params[4].validate == "numeric"



from squall import Squall
from squall.testclient import TestClient

app = Squall()


@app.put("/items/{item_id}")
def save_item_no_body(item_id: str):
    return {"item_id": item_id}


client = TestClient(app)


openapi_schema = {
    "openapi": "3.0.2",
    "info": {"title": "Squall", "version": "0.1.0"},
    "paths": {
        "/items/{item_id}": {
            "put": {
                "summary": "Save Item No Body",
                "operationId": "save_item_no_body_items__item_id__put",
                "parameters": [
                    {
                        "required": True,
                        "schema": {"type": "string"},
                        "name": "item_id",
                        "in": "path",
                    }
                ],
                "responses": {
                    "200": {
                        "description": "Successful Response",
                        "content": {"application/json": {"schema": {}}},
                    },
                    "400": {
                        "description": "Parameters Validation Error",
                        "content": {
                            "application/json": {
                                "schema": {
                                    "$ref": "#/components/schemas/HTTPBadRequestError"
                                }
                            }
                        },
                    },
                },
            }
        }
    },
}


def test_openapi_schema():
    response = client.get("/openapi.json")
    assert response.status_code == 200, response.text
    assert response.json() == openapi_schema


def test_put_no_body():
    response = client.put("/items/foo")
    assert response.status_code == 200, response.text
    assert response.json() == {"item_id": "foo"}


def test_put_no_body_with_body():
    response = client.put("/items/foo", json={"name": "Foo"})
    assert response.status_code == 200, response.text
    assert response.json() == {"item_id": "foo"}



from dataclasses import dataclass

from squall import Body, Squall
from squall.testclient import TestClient

app = Squall()

media_type = "application/vnd.api+json"


# NOTE: These are not valid JSON:API resources
# but they are fine for testing requestBody with custom media_type
@dataclass
class Product:
    name: str
    price: float


@app.post("/products")
async def create_product(data: Product = Body(..., media_type=media_type, embed=True)):
    pass  # pragma: no cover


create_product_request_body = {
    "content": {
        "application/vnd.api+json": {"schema": {"$ref": "#/components/schemas/Product"}}
    },
    "required": None,
}

client = TestClient(app)


def test_openapi_schema():
    response = client.get("/openapi.json")
    assert response.status_code == 200, response.text
    openapi_schema = response.json()
    assert (
        openapi_schema["paths"]["/products"]["post"]["requestBody"]
        == create_product_request_body
    )
